#ifndef _ASSING_MULT_HOP_EO_MIC_H
#define _ASSING_MULT_HOP_EO_MIC_H

#undef _COMM_OVERLAP_

#define _HOP_T
#define _HOP_Z
#define _HOP_Y
#define _HOP_X
#define _HOP_CLV

#if _NDIMZ != 1
#define _ZPAR   1
#else
#define _ZPAR   0
#endif
#if _NDIMY != 1
#define _YPAR   1
#else
#define _YPAR   0
#endif
#if _NDIMX != 1
#define _XPAR   1
#else
#define _XPAR   0
#endif

#define _PAR_DIR   ((_ZPAR) + (_YPAR)*2 + (_XPAR)*4)



extern "C" void
#ifdef _CLOVER
assign_mult_hop_eo_mic(      mic_wqf_eo      *restrict mye, 
                       const mic_wqf_eo      *restrict  yo,
                       const mic_gvf_eo_soa  *restrict  ue,
                       const mic_gvf_eo_soa  *restrict  uo,
                       const mic_clvf_eo      *fclinve) 
#else
assign_mult_hop_eo_mic(      mic_wqf_eo      *restrict mye, 
                       const mic_wqf_eo      *restrict  yo,
                       const mic_gvf_eo_soa  *restrict  ue,
                       const mic_gvf_eo_soa  *restrict  uo)
#endif
////////////////////////////////////////////////////////////////////////////////
// Multiply hopping matrix on the fermion field
//
//  mye = Fee Meo yo
//
//  mye : even/odd site fermion field (output)
//   yo : odd/even site fermion field (input)
//   ue : even/odd site gauge field (input)
//   uo : odd/even site gauge field (input)
// fclinve : even/odd site inverse clover term (input)
//
////////////////////////////////////////////////////////////////////////////////
{

//printf("Mult: mye %p yo %p ue %p uo %p fclinve %p\n",
//mye->s,yo->s,ue->s,uo->s,fclinve->s);
  __assume_aligned((void *)mye, _ALIGN_SIZE);
  __assume_aligned((void *) yo, _ALIGN_SIZE);
  __assume_aligned((void *) ue, _ALIGN_SIZE);
  __assume_aligned((void *) uo, _ALIGN_SIZE);
#ifdef _CLOVER
  __assume_aligned((void *)fclinve, _ALIGN_SIZE);
#endif

  startTimer(solver_timer.mult_w_comm_timer);

  int ipeo = mic_ipeo;
  int ieo  = (*ue).ieo;
  (*mye).ieo  = ieo;

  /////////////////
  // receive start 
  /////////////////
#ifndef _singlePU
  startTimer(solver_timer.ycopy_timer);
#if _NDIMZ != 1
  mic_combuf_recv_start(br_z_up, NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), mic_nodeid_z_up);
  mic_combuf_recv_start(br_z_dn, NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), mic_nodeid_z_dn);
#endif
#if _NDIMY != 1
  mic_combuf_recv_start(br_y_up, NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), mic_nodeid_y_up);
  mic_combuf_recv_start(br_y_dn, NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), mic_nodeid_y_dn);
#endif
#if _NDIMX != 1
  mic_combuf_recv_start(br_x_up, NRAS_GHOST_X*sizeof(__mic_half_spinor_t), mic_nodeid_x_up);
  mic_combuf_recv_start(br_x_dn, NRAS_GHOST_X*sizeof(__mic_half_spinor_t), mic_nodeid_x_dn);
#endif
  stopTimer(solver_timer.ycopy_timer);
#endif


#ifndef _singlePU
/////////////////////////////////////////////////////////////
// mult on surface sites 
// (pre compuptation, packing and buffering for send)
/////////////////////////////////////////////////////////////
  startTimer(solver_timer.mult_timer);
  startTimer(solver_timer.mult_pre_timer);
  {

    const int tilet = 4;
    const int tilex = 4, tiley = 4, tilez = 4;
    Percore cores[maxCores];

#pragma omp parallel
    {

#if _NDIMZ != 1
      int z_nblocksx = (NX   + tilex - 1) / tilex;
      int z_nblocksy = (NY   + tiley - 1) / tiley;
      int z_nblocksz = 2;
      int z_nblockst = (NTHT + tilet - 1) / tilet;
      int z_nblocks  = z_nblocksx * z_nblocksy * z_nblocksz * z_nblockst;
#else
      int z_nblocks  = 0;
#endif
#if _NDIMY != 1
      int y_nblocksx = (NX   + tilex - 1) / tilex;
      int y_nblocksy = 2;
      int y_nblocksz = (NZ   + tilez - 1) / tilez;
      int y_nblockst = (NTHT + tilet - 1) / tilet;
      int y_nblocks  = y_nblocksx * y_nblocksy * y_nblocksz * y_nblockst;
#else
      int y_nblocks  = 0;
#endif
#if _NDIMX != 1
      int x_nblocksx = 2;
      int x_nblocksy = (NY   + tiley - 1) / tiley;
      int x_nblocksz = (NZ   + tilez - 1) / tilez;
      int x_nblockst = (NTHT + tilet - 1) / tilet;
      int x_nblocks  = x_nblocksx * x_nblocksy * x_nblocksz * x_nblockst;
#else
      int x_nblocks  = 0;
#endif
      int nblocks = z_nblocks + y_nblocks + x_nblocks;

      Sched sch(nblocks, cores);
      #pragma omp barrier

      int block; 
      while ((block = sch.nextiter()) != -1)
      {

        struct offs {
          int isite_ctr;
          int isite_upt, isite_dnt;
          int isite_upz, isite_dnz;
          int isite_upy, isite_dny;
          int isite_upx, isite_dnx;
        } offsets[tilex*tiley*tilez*tilet/nHT+1];


        if (block < z_nblocks) {

#if _NDIMZ != 1
          /////////////
          // Z-send
          /////////////

          int jblock = block;

          int iix,iiy,iiz,iit,rem;
          int xmax,ymax,zmax,tmax;

          iix = jblock / (z_nblocksy*z_nblocksz*z_nblockst);
          rem = jblock % (z_nblocksy*z_nblocksz*z_nblockst);
          iiy =  rem   / (           z_nblocksz*z_nblockst);
          rem =  rem   % (           z_nblocksz*z_nblockst);
          iiz =  rem   / (                      z_nblockst);
          iit =  rem   % (                      z_nblockst);

          iix *= tilex;
          iiy *= tiley;
          iit *= tilet;
          xmax = iix + tilex; if (xmax > NX  ) xmax = NX;
          ymax = iiy + tiley; if (ymax > NY  ) ymax = NY;
          zmax = iiz;
          tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;

          int ioffs = 0;
          for (int ix = iix  ; ix < xmax  ; ++ix)
          for (int iy = iiy  ; iy < ymax  ; ++iy)
          for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
            int iz = iiz*(NZ-1);
            offsets[ioffs].isite_ctr = get_site(itht, iz, iy, ix);
            ++ioffs;
          }

          offsets[ioffs] = offsets[ioffs-1];

          ////////////////
          // Z-forward
          ////////////////
          if ( 0 == iiz ) {

            ioffs = 0;
            for (int ix = iix  ; ix < xmax  ; ++ix)
            for (int iy = iiy  ; iy < ymax  ; ++iy)
            for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
              const offs &cur_offs = offsets[ioffs];
              const offs &pf_offs  = offsets[ioffs+1];
              ++ioffs;
              const mic_su3_spinor *p_yo = (*yo).s;
              const mic_gsf_eo_soa *p_uo = (*uo).s;

              register __mic_half_spinor_t  w1,w2,w3;
  //            const mic_su3_spinor *yptr = &((*yo).s[cur_offs.isite_ctr]);
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[5], _MM_HINT_T1);
              __mic_mult_proj_pre_forw_z(w,yptr);

              int ibuff = COL*(itht + NTHT*(iy + NY*ix));
              _my_ngostore_ps(&(sbuff_z_dn[ibuff+0]),w1);
              _my_ngostore_ps(&(sbuff_z_dn[ibuff+1]),w2);
              _my_ngostore_ps(&(sbuff_z_dn[ibuff+2]),w3);

            }

          } else if ( 1 == iiz ) {
          ////////////////
          // Z-backward
          ////////////////

            ioffs = 0;
            for (int ix = iix  ; ix < xmax  ; ++ix)
            for (int iy = iiy  ; iy < ymax  ; ++iy)
            for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
              const offs &cur_offs = offsets[ioffs];
              const offs &pf_offs  = offsets[ioffs+1];
              ++ioffs;
              const mic_su3_spinor *p_yo = (*yo).s;
              const mic_gsf_eo_soa *p_uo = (*uo).s;

              register __mic_half_spinor_t  w1,w2,w3;
  //            const mic_su3_spinor *yptr = &((*yo).s[cur_offs.isite_ctr]);
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
              _my_pf(&p_uo[cur_offs.isite_ctr].v[2].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_ctr].v[2].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_ctr].v[2].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_z(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
  //              const mic_su3mat *uptr = &((*uo).s[cur_offs.isite_ctr].v[2]);
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_ctr].v[2];
                _my_pf(&p_uo[pf_offs.isite_ctr].v[2].u[0], _MM_HINT_T1);
                _my_pf(&p_uo[pf_offs.isite_ctr].v[2].u[1], _MM_HINT_T1);
                _my_pf(&p_uo[pf_offs.isite_ctr].v[2].u[2], _MM_HINT_T1);
                __mic_su3_load_u(u,uptr);

                _my_pf(&p_yo[ pf_offs.isite_ctr].y[0], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[1], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[2], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[3], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[4], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[5], _MM_HINT_T1);

                __mic_su3_gmultdag_vec(u,w);
              }

              int ibuff = COL*(itht + NTHT*(iy + NY*ix));
              _my_ngostore_ps(&(sbuff_z_up[ibuff+0]),w1);
              _my_ngostore_ps(&(sbuff_z_up[ibuff+1]),w2);
              _my_ngostore_ps(&(sbuff_z_up[ibuff+2]),w3);

            }

          } // end of if (iiz)
#endif

        } else if (z_nblocks <= block && block < z_nblocks + y_nblocks) {

#if _NDIMY != 1
          /////////////
          // Y-send
          /////////////

          int jblock = block - z_nblocks;

          int iix,iiy,iiz,iit,rem;
          int xmax,ymax,zmax,tmax;

          iix = jblock / (y_nblocksy*y_nblocksz*y_nblockst);
          rem = jblock % (y_nblocksy*y_nblocksz*y_nblockst);
          iiy =  rem   / (           y_nblocksz*y_nblockst);
          rem =  rem   % (           y_nblocksz*y_nblockst);
          iiz =  rem   / (                      y_nblockst);
          iit =  rem   % (                      y_nblockst);

          iix *= tilex;
          iiz *= tilez;
          iit *= tilet;
          xmax = iix + tilex; if (xmax > NX  ) xmax = NX;
          ymax = iiy;
          zmax = iiz + tilez; if (zmax > NZ  ) zmax = NZ;
          tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;

          int ioffs = 0;
          for (int ix = iix  ; ix < xmax  ; ++ix)
          for (int iz = iiz  ; iz < zmax  ; ++iz)
          for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
            int iy = iiy*(NY-1);
            offsets[ioffs].isite_ctr = get_site(itht, iz, iy, ix);
            ++ioffs;
          }

          offsets[ioffs] = offsets[ioffs-1];

          ////////////////
          // Y-forward
          ////////////////
          if ( 0 == iiy ) {

            ioffs = 0;
            for (int ix = iix  ; ix < xmax  ; ++ix)
            for (int iz = iiz  ; iz < zmax  ; ++iz)
            for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
              const offs &cur_offs = offsets[ioffs];
              const offs &pf_offs  = offsets[ioffs+1];
              ++ioffs;
              const mic_su3_spinor *p_yo = (*yo).s;
              const mic_gsf_eo_soa *p_uo = (*uo).s;

              register __mic_half_spinor_t  w1,w2,w3;
  //            const mic_su3_spinor *yptr = &((*yo).s[cur_offs.isite_ctr]);
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[5], _MM_HINT_T1);
              __mic_mult_proj_pre_forw_y(w,yptr);

              int ibuff = COL*(itht + NTHT*(iz + NZ*ix));
              _my_ngostore_ps(&(sbuff_y_dn[ibuff+0]),w1);
              _my_ngostore_ps(&(sbuff_y_dn[ibuff+1]),w2);
              _my_ngostore_ps(&(sbuff_y_dn[ibuff+2]),w3);

            }

          } else if ( 1 == iiy ) {
          ////////////////
          // Y-backward
          ////////////////

            ioffs = 0;
            for (int ix = iix  ; ix < xmax  ; ++ix)
            for (int iz = iiz  ; iz < zmax  ; ++iz)
            for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
              const offs &cur_offs = offsets[ioffs];
              const offs &pf_offs  = offsets[ioffs+1];
              ++ioffs;
              const mic_su3_spinor *p_yo = (*yo).s;
              const mic_gsf_eo_soa *p_uo = (*uo).s;

              register __mic_half_spinor_t  w1,w2,w3;
  //            const mic_su3_spinor *yptr = &((*yo).s[cur_offs.isite_ctr]);
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
              _my_pf(&p_uo[cur_offs.isite_ctr].v[1].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_ctr].v[1].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_ctr].v[1].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_y(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
  //              const mic_su3mat *uptr = &((*uo).s[cur_offs.isite_ctr].v[1]);
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_ctr].v[1];
                _my_pf(&p_uo[pf_offs.isite_ctr].v[1].u[0], _MM_HINT_T1);
                _my_pf(&p_uo[pf_offs.isite_ctr].v[1].u[1], _MM_HINT_T1);
                _my_pf(&p_uo[pf_offs.isite_ctr].v[1].u[2], _MM_HINT_T1);
                __mic_su3_load_u(u,uptr);

                _my_pf(&p_yo[ pf_offs.isite_ctr].y[0], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[1], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[2], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[3], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[4], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[5], _MM_HINT_T1);

                __mic_su3_gmultdag_vec(u,w);
              }

              int ibuff = COL*(itht + NTHT*(iz + NZ*ix));
              _my_ngostore_ps(&(sbuff_y_up[ibuff+0]),w1);
              _my_ngostore_ps(&(sbuff_y_up[ibuff+1]),w2);
              _my_ngostore_ps(&(sbuff_y_up[ibuff+2]),w3);

            }

          } // end of if (iiy)
#endif

        } else if (z_nblocks + y_nblocks <= block && block < z_nblocks + y_nblocks + x_nblocks) {

#if _NDIMX != 1
          /////////////
          // X-send
          /////////////

          int jblock = block - z_nblocks - y_nblocks;

          int iix,iiy,iiz,iit,rem;
          int xmax,ymax,zmax,tmax;

          iix = jblock / (x_nblocksy*x_nblocksz*x_nblockst);
          rem = jblock % (x_nblocksy*x_nblocksz*x_nblockst);
          iiy =  rem   / (           x_nblocksz*x_nblockst);
          rem =  rem   % (           x_nblocksz*x_nblockst);
          iiz =  rem   / (                      x_nblockst);
          iit =  rem   % (                      x_nblockst);

          iiy *= tiley;
          iiz *= tilez;
          iit *= tilet;
          xmax = iix;
          ymax = iiy + tiley; if (ymax > NY  ) ymax = NY;
          zmax = iiz + tilez; if (zmax > NZ  ) zmax = NZ;
          tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;

          int ioffs = 0;
          for (int iy = iiy  ; iy < ymax  ; ++iy)
          for (int iz = iiz  ; iz < zmax  ; ++iz)
          for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
            int ix = iix*(NX-1);
            offsets[ioffs].isite_ctr = get_site(itht, iz, iy, ix);
            ++ioffs;
          }
          offsets[ioffs] = offsets[ioffs-1];

          ////////////////
          // X-forward
          ////////////////
          if ( 0 == iix ) {

            ioffs = 0;
            for (int iy = iiy  ; iy < ymax  ; ++iy)
            for (int iz = iiz  ; iz < zmax  ; ++iz)
            for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
              const offs &cur_offs = offsets[ioffs];
              const offs &pf_offs  = offsets[ioffs+1];
              ++ioffs;
              const mic_su3_spinor *p_yo = (*yo).s;
              const mic_gsf_eo_soa *p_uo = (*uo).s;

              register __mic_half_spinor_t  w1,w2,w3;
  //            const mic_su3_spinor *yptr = &((*yo).s[cur_offs.isite_ctr]);
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_ctr].y[5], _MM_HINT_T1);
              __mic_mult_proj_pre_forw_x(w,yptr);

              int ibuff = COL*(itht + NTHT*(iy + NY*iz));
              _my_ngostore_ps(&(sbuff_x_dn[ibuff+0]),w1);
              _my_ngostore_ps(&(sbuff_x_dn[ibuff+1]),w2);
              _my_ngostore_ps(&(sbuff_x_dn[ibuff+2]),w3);

            }

          } else if ( 1 == iix ) {
          ////////////////
          // X-backward
          ////////////////

            ioffs = 0;
            for (int iy = iiy  ; iy < ymax  ; ++iy)
            for (int iz = iiz  ; iz < zmax  ; ++iz)
            for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
              const offs &cur_offs = offsets[ioffs];
              const offs &pf_offs  = offsets[ioffs+1];
              ++ioffs;
              const mic_su3_spinor *p_yo = (*yo).s;
              const mic_gsf_eo_soa *p_uo = (*uo).s;

              register __mic_half_spinor_t  w1,w2,w3;
  //            const mic_su3_spinor *yptr = &((*yo).s[cur_offs.isite_ctr]);
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
              _my_pf(&p_uo[cur_offs.isite_ctr].v[0].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_ctr].v[0].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_ctr].v[0].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_x(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
  //              const mic_su3mat *uptr = &((*uo).s[cur_offs.isite_ctr].v[0]);
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_ctr].v[0];
                _my_pf(&p_uo[pf_offs.isite_ctr].v[0].u[0], _MM_HINT_T1);
                _my_pf(&p_uo[pf_offs.isite_ctr].v[0].u[1], _MM_HINT_T1);
                _my_pf(&p_uo[pf_offs.isite_ctr].v[0].u[2], _MM_HINT_T1);
                __mic_su3_load_u(u,uptr);

                _my_pf(&p_yo[ pf_offs.isite_ctr].y[0], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[1], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[2], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[3], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[4], _MM_HINT_T1);
                _my_pf(&p_yo[ pf_offs.isite_ctr].y[5], _MM_HINT_T1);

                __mic_su3_gmultdag_vec(u,w);
              }

              int ibuff = COL*(itht + NTHT*(iy + NY*iz));
              _my_ngostore_ps(&(sbuff_x_up[ibuff+0]),w1);
              _my_ngostore_ps(&(sbuff_x_up[ibuff+1]),w2);
              _my_ngostore_ps(&(sbuff_x_up[ibuff+2]),w3);

            }

          } // end of if (iix)
#endif

        }

      } // end of while block
    } // end of omp parallel
  }
  stopTimer(solver_timer.mult_pre_timer);
  stopTimer(solver_timer.mult_timer);
#endif

////////////////////////////
// send 
////////////////////////////
#ifndef _singlePU
  {
    startTimer(solver_timer.ycopy_timer);
#if _NDIMZ != 1
    mic_combuf_send_start(bs_z_dn, NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), mic_nodeid_z_dn);
    mic_combuf_send_start(bs_z_up, NRAS_GHOST_Z*sizeof(__mic_half_spinor_t), mic_nodeid_z_up);
#endif
#if _NDIMY != 1
    mic_combuf_send_start(bs_y_dn, NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), mic_nodeid_y_dn);
    mic_combuf_send_start(bs_y_up, NRAS_GHOST_Y*sizeof(__mic_half_spinor_t), mic_nodeid_y_up);
#endif
#if _NDIMX != 1
    mic_combuf_send_start(bs_x_dn, NRAS_GHOST_X*sizeof(__mic_half_spinor_t), mic_nodeid_x_dn);
    mic_combuf_send_start(bs_x_up, NRAS_GHOST_X*sizeof(__mic_half_spinor_t), mic_nodeid_x_up);
#endif
    stopTimer(solver_timer.ycopy_timer);
  }
#endif

///////////////////////////////////
// mult on bulk sites
///////////////////////////////////
  startTimer(solver_timer.mult_timer);
#ifndef _singlePU
  startTimer(solver_timer.mult_in_timer);
#endif
  {

    const int tilex = 2, tiley = 2, tilez = 4, tilet = 8;
    Percore cores[maxCores];
#if _NDIMX != 1 
    const int nblocksx = (NX-2 + tilex - 1) / tilex;
#else 
    const int nblocksx = (NX   + tilex - 1) / tilex;
#endif
#if _NDIMY != 1
    const int nblocksy = (NY-2 + tiley - 1) / tiley; 
#else 
    const int nblocksy = (NY   + tiley - 1) / tiley; 
#endif
#if _NDIMZ != 1
    const int nblocksz = (NZ-2 + tilez - 1) / tilez;
#else 
    const int nblocksz = (NZ   + tilez - 1) / tilez;
#endif
    const int nblockst = (NTHT + tilet - 1) / tilet;
    const int nblocks = nblocksx * nblocksy * nblocksz * nblockst;
    const int offsize = nblocks * tilex * tiley * tilez * tilet;

    static struct offs {
      unsigned isite_ctr;
      unsigned isite_upt, isite_dnt;
      unsigned isite_upz, isite_dnz;
      unsigned isite_upy, isite_dny;
      unsigned isite_upx, isite_dnx;
    } offsets[offsize+1];
    static bool offsinit = false;
    if (!offsinit) {
      offsinit = true;
      #pragma omp parallel
    {
      int nblockst = (NTHT + tilet - 1) / tilet;
      int nblocks = nblocksx * nblocksy * nblocksz * nblockst;
      Sched sch(nblocks, cores);
      #pragma omp barrier
      int block;
      while ((block = sch.nextiter()) != -1)
      {
        int iix = block / (nblocksy*nblocksz*nblockst);
        int rem = block % (nblocksy*nblocksz*nblockst);
        int iiy = rem   / (         nblocksz*nblockst);
        rem     = rem   % (         nblocksz*nblockst);
        int iiz = rem   / (                  nblockst);
        int iit = rem   % (                  nblockst);

        int ioffs = (iit + nblockst*(iiz + nblocksz*(iiy + nblocksy*iix)));
        ioffs *= tilex * tiley * tilez * tilet;

        iiz *= tilez;
        iiy *= tiley;
        iix *= tilex;
        iit *= tilet;
#if _NDIMZ != 1
        int zmax = iiz + tilez; if (zmax > NZ-2) zmax = NZ-2;
#else
        int zmax = iiz + tilez; if (zmax > NZ  ) zmax = NZ;
#endif
#if _NDIMY != 1
        int ymax = iiy + tiley; if (ymax > NY-2) ymax = NY-2;
#else
        int ymax = iiy + tiley; if (ymax > NY  ) ymax = NY;
#endif
#if _NDIMX != 1
        int xmax = iix + tilex; if (xmax > NX-2) xmax = NX-2;
#else
        int xmax = iix + tilex; if (xmax > NX  ) xmax = NX;
#endif
        int tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;



#if _NDIMX != 1
        for (int ix = iix+1; ix < xmax+1; ++ix)
#else
        for (int ix = iix  ; ix < xmax  ; ++ix)
#endif
#if _NDIMY != 1
        for (int iy = iiy+1; iy < ymax+1; ++iy)
#else
        for (int iy = iiy  ; iy < ymax  ; ++iy)
#endif
#if _NDIMZ != 1
        for (int iz = iiz+1; iz < zmax+1; ++iz)
#else
        for (int iz = iiz  ; iz < zmax  ; ++iz)
#endif
        for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
          int ioffs1 = ioffs + sch.myht;
          if (ioffs1 >= offsize)
          {
            fprintf(stderr, "Error: overflow: ioffs %d >= %d\n",
              ioffs, offsize);
            exit(1);
          }
          offsets[ioffs1].isite_ctr = get_site(itht, iz, iy, ix);
          offsets[ioffs1].isite_upt = get_site((itht+1)%NTHT,iz,iy,ix);
          offsets[ioffs1].isite_dnt = get_site((itht-1+NTHT)%NTHT,iz,iy,ix);
          offsets[ioffs1].isite_upz = get_site(itht,(iz+1)%NZ,iy,ix);
          offsets[ioffs1].isite_dnz = get_site(itht,(iz-1+NZ)%NZ,iy,ix);
          offsets[ioffs1].isite_upy = get_site(itht,iz,(iy+1)%NY,ix);
          offsets[ioffs1].isite_dny = get_site(itht,iz,(iy-1+NY)%NY,ix);
          offsets[ioffs1].isite_upx = get_site(itht,iz,iy,(ix+1)%NX);
          offsets[ioffs1].isite_dnx = get_site(itht,iz,iy,(ix-1+NX)%NX);
          ioffs += nHT;
        }
      }
    } // parallel
    offsets[offsize] = offsets[offsize-1];
    }

    #ifdef PMU_STAT
    solver_timer.mult_stat.start();
    #endif
#pragma omp parallel
    {
#if _NDIMX != 1
      int nblocksx = (NX-2 + tilex - 1) / tilex;
#else
      int nblocksx = (NX   + tilex - 1) / tilex;
#endif
#if _NDIMY != 1
      int nblocksy = (NY-2 + tiley - 1) / tiley;
#else
      int nblocksy = (NY   + tiley - 1) / tiley;
#endif
#if _NDIMZ != 1
      int nblocksz = (NZ-2 + tilez - 1) / tilez;
#else
      int nblocksz = (NZ   + tilez - 1) / tilez;
#endif
      int nblockst = (NTHT + tilet - 1) / tilet;
      int nblocks = nblocksx * nblocksy * nblocksz * nblockst;
      Sched sch(nblocks, cores);

      #pragma omp barrier
      #ifdef PMU_STAT
      solver_timer.mult_stat.enter(sch.mythread);
      #endif
      int block;
      while ((block = sch.nextiter()) != -1)
      {
        int iix = block / (nblocksy*nblocksz*nblockst);
        int rem = block % (nblocksy*nblocksz*nblockst);
        int iiy = rem   / (         nblocksz*nblockst);
        rem     = rem   % (         nblocksz*nblockst);
        int iiz = rem   / (                  nblockst);
        int iit = rem   % (                  nblockst);

        int ioffs = (iit + nblockst*(iiz + nblocksz*(iiy + nblocksy*iix)));
        ioffs *= tilex * tiley * tilez * tilet;

        iiz *= tilez;
        iiy *= tiley;
        iix *= tilex;
        iit *= tilet;
#if _NDIMZ != 1
        int zmax = iiz + tilez; if (zmax > NZ-2) zmax = NZ-2;
#else
        int zmax = iiz + tilez; if (zmax > NZ  ) zmax = NZ;
#endif
#if _NDIMY != 1
        int ymax = iiy + tiley; if (ymax > NY-2) ymax = NY-2;
#else
        int ymax = iiy + tiley; if (ymax > NY  ) ymax = NY;
#endif
#if _NDIMX != 1
        int xmax = iix + tilex; if (xmax > NX-2) xmax = NX-2;
#else
        int xmax = iix + tilex; if (xmax > NX  ) xmax = NX;
#endif
        int tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;


#if _NDIMX != 1
        for (int ix = iix+1; ix < xmax+1; ++ix)
#else
        for (int ix = iix  ; ix < xmax  ; ++ix)
#endif
#if _NDIMY != 1
        for (int iy = iiy+1; iy < ymax+1; ++iy)
#else
        for (int iy = iiy  ; iy < ymax  ; ++iy)
#endif
#if _NDIMZ != 1
        for (int iz = iiz+1; iz < zmax+1; ++iz)
#else
        for (int iz = iiz  ; iz < zmax  ; ++iz)
#endif
        for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {

          int ieoxyz = (ipeo + (ix+1) + (iy+1) + (iz+1) + ieo) % 2;
          int ioffs1 = ioffs + sch.myht;
          const offs &cur_offs = offsets[ioffs1];
          const offs &pf_offs = offsets[ioffs1+nHT];
          ioffs += nHT;
          const mic_gsf_eo_soa *p_ue = (*ue).s;
          const mic_gsf_eo_soa &r_ue = p_ue[cur_offs.isite_ctr];
          const mic_gsf_eo_soa *p_uo = (*uo).s;
          const mic_su3_spinor *p_yo = (*yo).s;
#ifdef _CLOVER
          const mic_clover *fptr = (*fclinve).s;
#endif

          register __mic_half_spinor_t yy1,yy2,yy3,yy4,yy5,yy6;
          {
            register __mic_half_spinor_t zero = _my_setzero_ps();

            yy1 = zero;
            yy2 = zero;
            yy3 = zero;
            yy4 = zero;
            yy5 = zero;
            yy6 = zero;
          }

#ifdef _HOP_T
          ////////////////////////////////////////
          // T-forward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;
            const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
            _my_pf(&r_ue.v[3].u[0], _MM_HINT_T0);
            _my_pf(&r_ue.v[3].u[1], _MM_HINT_T0);
            _my_pf(&r_ue.v[3].u[2], _MM_HINT_T0);
            __mic_mult_proj_pre_forw_t(w,yptr);

            if ( 0 == ieoxyz ) {

              register __mic_half_spinor_t  s1,s2,s3;

              // periodic boundary condition in T-direction

              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upt];
              __mic_mult_proj_pre_forw_t(s,yptr);

              __mic_mult_shift_forw_t(w,s);
            }

            register __mic_half_spinor_t  u1,u2,u3;
            const mic_su3mat *uptr = &r_ue.v[3];
            _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[0], _MM_HINT_T2);
            _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[1], _MM_HINT_T2);
            _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[2], _MM_HINT_T2);
            __mic_su3_load_u(u,uptr);

            __mic_su3_gmult_vec(u,w);

            __mic_mult_proj_post_forw_t(yy,w);

          }

          ////////////////////////////////////////
          // T-backward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;
            const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
            _my_pf(&p_uo[cur_offs.isite_ctr].v[3].u[0], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_ctr].v[3].u[1], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_ctr].v[3].u[2], _MM_HINT_T0);
            __mic_mult_proj_pre_back_t(w,yptr);

            register __mic_half_spinor_t  u1,u2,u3;
            const mic_su3mat *uptr = &p_uo[cur_offs.isite_ctr].v[3];
            _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[0], _MM_HINT_T1);
            _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[1], _MM_HINT_T1);
            _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[2], _MM_HINT_T1);
            _my_pf(&p_uo[cur_offs.isite_dnt].v[3].u[0], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_dnt].v[3].u[1], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_dnt].v[3].u[2], _MM_HINT_T0);
            __mic_su3_load_u(u,uptr);

            if ( 1 == ieoxyz ) {

              // periodic boundary condition in T-direction
              register __mic_half_spinor_t  s1,s2,s3;

              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnt];
              __mic_mult_proj_pre_back_t(s,yptr);

              __mic_mult_shift_back_t(w,s);

              const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnt].v[3];
              __mic_su3_load_u(s,uptr);

              __mic_mult_shift_back_t(u,s);

            }

            _my_pf(&p_yo[cur_offs.isite_upz].y[0], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[1], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[2], _MM_HINT_T0);
            __mic_su3_gmultdag_vec(u,w);

            _my_pf(&p_yo[cur_offs.isite_upz].y[3], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[4], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[5], _MM_HINT_T0);
            __mic_mult_proj_post_back_t(yy,w);

          }
#endif

#ifdef _HOP_Z
          ////////////////////////////////////////
          // Z-forward
          ////////////////////////////////////////
          {

            register __mic_half_spinor_t  w1,w2,w3;

            {

              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upz];
              _my_pf(&p_yo[ pf_offs.isite_upz].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[5], _MM_HINT_T1);
              _my_pf(&r_ue.v[2].u[0], _MM_HINT_T0);
              _my_pf(&r_ue.v[2].u[1], _MM_HINT_T0);
              _my_pf(&r_ue.v[2].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_forw_z(w,yptr);

            }

            {
              register __mic_half_spinor_t  u1,u2,u3;
              const mic_su3mat *uptr = &r_ue.v[2];
              _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[0], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[1], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[2], _MM_HINT_T2);
              __mic_su3_load_u(u,uptr);

              _my_pf(&p_yo[cur_offs.isite_dnz].y[0], _MM_HINT_T0);
              _my_pf(&p_yo[cur_offs.isite_dnz].y[1], _MM_HINT_T0);
              _my_pf(&p_yo[cur_offs.isite_dnz].y[2], _MM_HINT_T0);
              __mic_su3_gmult_vec(u,w);
            }

            _my_pf(&p_yo[cur_offs.isite_dnz].y[3], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_dnz].y[4], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_dnz].y[5], _MM_HINT_T0);
            __mic_mult_proj_post_forw_z(yy,w);

          }

          ////////////////////////////////////////
          // Z-backward
          ////////////////////////////////////////
          {

            register __mic_half_spinor_t  w1,w2,w3;

            {
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnz];
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[5], _MM_HINT_T1);
              _my_pf(&p_uo[cur_offs.isite_dnz].v[2].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnz].v[2].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnz].v[2].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_z(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnz].v[2];
                _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[0], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[1], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[2], _MM_HINT_T2);
                __mic_su3_load_u(u,uptr);

                __mic_su3_gmultdag_vec(u,w);
              }
            }

            __mic_mult_proj_post_back_z(yy,w);
          }
#endif

#ifdef _HOP_Y
          ////////////////////////////////////////
          // Y-forward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;

            {
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upy];
              _my_pf(&p_yo[ pf_offs.isite_upy].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[5], _MM_HINT_T1);
              _my_pf(&r_ue.v[1].u[0], _MM_HINT_T0);
              _my_pf(&r_ue.v[1].u[1], _MM_HINT_T0);
              _my_pf(&r_ue.v[1].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_forw_y(w,yptr);
            }

            {
              const mic_su3mat *uptr = &r_ue.v[1];
              _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[0], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[1], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[2], _MM_HINT_T2);

              register __mic_half_spinor_t  u1,u2,u3;
              __mic_su3_load_u(u,uptr);

              __mic_su3_gmult_vec(u,w);
            }

            __mic_mult_proj_post_forw_y(yy,w);

          }

          ////////////////////////////////////////
          // Y-backward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;
            {
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dny];
              _my_pf(&p_yo[ pf_offs.isite_dny].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[5], _MM_HINT_T1);
              _my_pf(&p_uo[cur_offs.isite_dny].v[1].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dny].v[1].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dny].v[1].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_y(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_dny].v[1];
                _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[0], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[1], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[2], _MM_HINT_T2);
                __mic_su3_load_u(u,uptr);

                __mic_su3_gmultdag_vec(u,w);
              }
            }

            __mic_mult_proj_post_back_y(yy,w);

          }
#endif

#ifdef _HOP_X
          ////////////////////////////////////////
          // X-forward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;

            {
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upx];
              _my_pf(&p_yo[ pf_offs.isite_upx].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[5], _MM_HINT_T1);
              _my_pf(&r_ue.v[0].u[0], _MM_HINT_T0);
              _my_pf(&r_ue.v[0].u[1], _MM_HINT_T0);
              _my_pf(&r_ue.v[0].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_forw_x(w,yptr);
            }

            {
              register __mic_half_spinor_t  u1,u2,u3;

              const mic_su3mat *uptr = &r_ue.v[0];
              _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[0], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[1], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[2], _MM_HINT_T2);
              __mic_su3_load_u(u,uptr);

              __mic_su3_gmult_vec(u,w);
            }

            __mic_mult_proj_post_forw_x(yy,w);

          }

          ////////////////////////////////////////
          // X-backward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;

            {
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnx];
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[5], _MM_HINT_T1);
              _my_pf(&p_uo[cur_offs.isite_dnx].v[0].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnx].v[0].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnx].v[0].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_x(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnx].v[0];
                _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[0], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[1], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[2], _MM_HINT_T2);
                __mic_su3_load_u(u,uptr);

                __mic_su3_gmultdag_vec(u,w);
              }
            }
            __mic_mult_proj_post_back_x(yy,w);

          }
#endif

#ifdef _HOP_CLV
#ifdef _CLOVER
          {
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[0], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[1], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[2], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[3], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[4], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[5], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[6], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[7], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[8], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[0], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[1], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[2], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[3], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[4], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[5], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[6], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[7], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[8], _MM_HINT_T2);
            __mic_mult_clvoer_term(yy,&fptr[cur_offs.isite_ctr]);
          }
#endif
#endif

          {
            __mic_half_spinor_t *ye = &(*mye).s[cur_offs.isite_ctr].y[0];
            _my_ngostore_ps(&ye[0],yy1);
            _my_ngostore_ps(&ye[1],yy2);
            _my_ngostore_ps(&ye[2],yy3);
            _my_ngostore_ps(&ye[3],yy4);
            _my_ngostore_ps(&ye[4],yy5);
            _my_ngostore_ps(&ye[5],yy6);
          }

        }

      } // end of while
      #ifdef PMU_STAT
      solver_timer.mult_stat.exit(sch.mythread);
      #endif
    }   // end of omp parallel
  }
  #ifdef PMU_STAT
  solver_timer.mult_stat.accum(omp_get_max_threads());
  #endif
#ifndef _singlePU
  stopTimer(solver_timer.mult_in_timer);
#endif
  stopTimer(solver_timer.mult_timer);

////////////////////////////
// wait for recv
////////////////////////////
#ifndef _singlePU
  startTimer(solver_timer.ycopy_timer);
  startTimer(solver_timer.ycopy_wait_timer);
  mic_combuf_waitall();
  stopTimer(solver_timer.ycopy_wait_timer);
  stopTimer(solver_timer.ycopy_timer);
#endif


#ifndef _singlePU
///////////////////////////////////////////////
// mult on surface sites (post compuptation)
///////////////////////////////////////////////
  startTimer(solver_timer.mult_timer);
  startTimer(solver_timer.mult_pst_timer);
  {

    const int tilet = 4;
    const int tilex = 4, tiley = 4, tilez = 4;
    Percore cores[maxCores];

#pragma omp parallel
    {

#if _NDIMZ != 1
      int z_nblocksx = (NX   + tilex - 1) / tilex;
      int z_nblocksy = (NY   + tiley - 1) / tiley;
      int z_nblocksz = 2;
      int z_nblockst = (NTHT + tilet - 1) / tilet;
      int z_nblocks  = z_nblocksx * z_nblocksy * z_nblocksz * z_nblockst;
#else
      int z_nblocks  = 0;
#endif
#if _NDIMY != 1
      int y_nblocksx = (NX   + tilex - 1) / tilex;
      int y_nblocksy = 2;
      int y_nblocksz = (NZ   + tilez - 1) / tilez;
      int y_nblockst = (NTHT + tilet - 1) / tilet;
      int y_nblocks  = y_nblocksx * y_nblocksy * y_nblocksz * y_nblockst;
#else
      int y_nblocks  = 0;
#endif
#if _NDIMX != 1
      int x_nblocksx = 2;
      int x_nblocksy = (NY   + tiley - 1) / tiley;
      int x_nblocksz = (NZ   + tilez - 1) / tilez;
      int x_nblockst = (NTHT + tilet - 1) / tilet;
      int x_nblocks  = x_nblocksx * x_nblocksy * x_nblocksz * x_nblockst;
#else
      int x_nblocks  = 0;
#endif
      int nblocks = z_nblocks + y_nblocks + x_nblocks;

      Sched sch(nblocks, cores);
      #pragma omp barrier

      int block; 
      while ((block = sch.nextiter()) != -1)
      {

        struct offs {
          int isite_ctr;
          int isite_upt, isite_dnt;
          int isite_upz, isite_dnz;
          int isite_upy, isite_dny;
          int isite_upx, isite_dnx;
          int ibuff_z;
          int ibuff_y;
          int ibuff_x;
        } offsets[tilex*tiley*tilez*tilet/nHT+1];

        int iix,iiy,iiz,iit,rem;
        int xmax,ymax,zmax,tmax;

        if (block < z_nblocks) {
#if _NDIMZ != 1
          int jblock = block;

          iix = jblock / (z_nblocksy*z_nblocksz*z_nblockst);
          rem = jblock % (z_nblocksy*z_nblocksz*z_nblockst);
          iiy =  rem   / (           z_nblocksz*z_nblockst);
          rem =  rem   % (           z_nblocksz*z_nblockst);
          iiz =  rem   / (                      z_nblockst);
          iit =  rem   % (                      z_nblockst);

          iix *= tilex;
          iiy *= tiley;
          iiz *= (NZ-1);
          iit *= tilet;
          xmax = iix + tilex; if (xmax > NX  ) xmax = NX;
          ymax = iiy + tiley; if (ymax > NY  ) ymax = NY;
          zmax = iiz + 1;
          tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;
#endif
        } else if (z_nblocks <= block && block < z_nblocks + y_nblocks) {
#if _NDIMY != 1
          int jblock = block - z_nblocks;

          iix = jblock / (y_nblocksy*y_nblocksz*y_nblockst);
          rem = jblock % (y_nblocksy*y_nblocksz*y_nblockst);
          iiy =  rem   / (           y_nblocksz*y_nblockst);
          rem =  rem   % (           y_nblocksz*y_nblockst);
          iiz =  rem   / (                      y_nblockst);
          iit =  rem   % (                      y_nblockst);

          iix *= tilex;
          iiy *= (NY-1);
          iiz *= tilez;
          iit *= tilet;
          xmax = iix + tilex; if (xmax > NX  ) xmax = NX;
          ymax = iiy + 1;
#if _NDIMZ != 1
          zmax = iiz + tilez; if (zmax > NZ-1) zmax = NZ-1; if (iiz == 0) iiz = 1;
#else
          zmax = iiz + tilez; if (zmax > NZ  ) zmax = NZ;
#endif
          tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;

#endif
        } else if (z_nblocks + y_nblocks <= block && block < z_nblocks + y_nblocks + x_nblocks) {
#if _NDIMX != 1
          int jblock = block - z_nblocks - y_nblocks;

          iix = jblock / (x_nblocksy*x_nblocksz*x_nblockst);
          rem = jblock % (x_nblocksy*x_nblocksz*x_nblockst);
          iiy =  rem   / (           x_nblocksz*x_nblockst);
          rem =  rem   % (           x_nblocksz*x_nblockst);
          iiz =  rem   / (                      x_nblockst);
          iit =  rem   % (                      x_nblockst);

          iix *= (NX-1);
          iiy *= tiley;
          iiz *= tilez;
          iit *= tilet;
          xmax = iix + 1;
#if _NDIMY != 1
          ymax = iiy + tiley; if (ymax > NY-1) ymax = NY-1; if (iiy == 0) iiy = 1;
#else
          ymax = iiy + tiley; if (ymax > NY  ) ymax = NY;
#endif
#if _NDIMZ != 1
          zmax = iiz + tilez; if (zmax > NZ-1) zmax = NZ-1; if (iiz == 0) iiz = 1;
#else
          zmax = iiz + tilez; if (zmax > NZ  ) zmax = NZ;
#endif
          tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;
#endif
        }

        int ioffs = 0;
        for (int ix = iix  ; ix < xmax  ; ++ix)
        for (int iy = iiy  ; iy < ymax  ; ++iy)
        for (int iz = iiz  ; iz < zmax  ; ++iz)
        for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
          offsets[ioffs].isite_ctr = get_site(itht, iz, iy, ix);
          offsets[ioffs].isite_upt = get_site((itht+1)%NTHT,iz,iy,ix);
          offsets[ioffs].isite_dnt = get_site((itht-1+NTHT)%NTHT,iz,iy,ix);
          offsets[ioffs].isite_upz = get_site(itht,(iz+1)%NZ,iy,ix);
          offsets[ioffs].isite_dnz = get_site(itht,(iz-1+NZ)%NZ,iy,ix);
          offsets[ioffs].isite_upy = get_site(itht,iz,(iy+1)%NY,ix);
          offsets[ioffs].isite_dny = get_site(itht,iz,(iy-1+NY)%NY,ix);
          offsets[ioffs].isite_upx = get_site(itht,iz,iy,(ix+1)%NX);
          offsets[ioffs].isite_dnx = get_site(itht,iz,iy,(ix-1+NX)%NX);
          offsets[ioffs].ibuff_z   = COL*(itht + NTHT*(iy + NY*ix));
          offsets[ioffs].ibuff_y   = COL*(itht + NTHT*(iz + NZ*ix));
          offsets[ioffs].ibuff_x   = COL*(itht + NTHT*(iy + NY*iz));
          ++ioffs;
        }

        offsets[ioffs] = offsets[ioffs-1];

        ioffs = 0;
        for (int ix = iix  ; ix < xmax  ; ++ix)
        for (int iy = iiy  ; iy < ymax  ; ++iy)
        for (int iz = iiz  ; iz < zmax  ; ++iz)
        for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {

          int ieoxyz = (ipeo + (ix+1) + (iy+1) + (iz+1) + ieo) % 2;
          const offs &cur_offs = offsets[ioffs];
          const offs &pf_offs  = offsets[ioffs+1];
          ++ioffs;
          const mic_gsf_eo_soa *p_ue = (*ue).s;
          const mic_gsf_eo_soa &r_ue = p_ue[cur_offs.isite_ctr];
          const mic_gsf_eo_soa *p_uo = (*uo).s;
          const mic_su3_spinor *p_yo = (*yo).s;
#ifdef _CLOVER
          const mic_clover *fptr = (*fclinve).s;
#endif

          register __mic_half_spinor_t yy1,yy2,yy3,yy4,yy5,yy6;
          {
            register __mic_half_spinor_t zero = _my_setzero_ps();

            yy1 = zero;
            yy2 = zero;
            yy3 = zero;
            yy4 = zero;
            yy5 = zero;
            yy6 = zero;
          }

#ifdef _HOP_T
          ////////////////////////////////////////
          // T-forward
          ////////////////////////////////////////
          {
            int isite = get_site(itht,iz,iy,ix);
            register __mic_half_spinor_t  w1,w2,w3;
            const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
            _my_pf(&r_ue.v[3].u[0], _MM_HINT_T0);
            _my_pf(&r_ue.v[3].u[1], _MM_HINT_T0);
            _my_pf(&r_ue.v[3].u[2], _MM_HINT_T0);
            __mic_mult_proj_pre_forw_t(w,yptr);

            if ( 0 == ieoxyz ) {

              // periodic boundary condition in T-direction
    //          int isite_up = (itht+1)%NTHT + NTHT*(iz + NZ*(iy + NY*ix));
              int isite_up = get_site((itht+1)%NTHT,iz,iy,ix);

              register __mic_half_spinor_t  s1,s2,s3;
              const mic_su3_spinor *yptr = &((*yo).s[isite_up]);
              __mic_mult_proj_pre_forw_t(s,yptr);

              __mic_mult_shift_forw_t(w,s);

            }

            register __mic_half_spinor_t  u1,u2,u3;
            const mic_su3mat *uptr = &r_ue.v[3];
            _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[0], _MM_HINT_T2);
            _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[1], _MM_HINT_T2);
            _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[2], _MM_HINT_T2);
            __mic_su3_load_u(u,uptr);

            __mic_su3_gmult_vec(u,w);

            __mic_mult_proj_post_forw_t(yy,w);

          }
          ////////////////////////////////////////
          // T-backward
          ////////////////////////////////////////
          {
            int isite = get_site(itht,iz,iy,ix);
            register __mic_half_spinor_t  w1,w2,w3;
            const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
            _my_pf(&p_uo[cur_offs.isite_ctr].v[3].u[0], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_ctr].v[3].u[1], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_ctr].v[3].u[2], _MM_HINT_T0);
            __mic_mult_proj_pre_back_t(w,yptr);

            register __mic_half_spinor_t  u1,u2,u3;
            const mic_su3mat *uptr = &p_uo[cur_offs.isite_ctr].v[3];
            _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[0], _MM_HINT_T1);
            _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[1], _MM_HINT_T1);
            _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[2], _MM_HINT_T1);
            _my_pf(&p_uo[cur_offs.isite_dnt].v[3].u[0], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_dnt].v[3].u[1], _MM_HINT_T0);
            _my_pf(&p_uo[cur_offs.isite_dnt].v[3].u[2], _MM_HINT_T0);
            __mic_su3_load_u(u,uptr);

            if ( 1 == ieoxyz ) {

              // periodic boundary condition in T-direction
              register __mic_half_spinor_t  s1,s2,s3;

              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnt];
              __mic_mult_proj_pre_back_t(s,yptr);

              __mic_mult_shift_back_t(w,s);

              const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnt].v[3];
              __mic_su3_load_u(s,uptr);

              __mic_mult_shift_back_t(u,s);

            }

            _my_pf(&p_yo[cur_offs.isite_upz].y[0], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[1], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[2], _MM_HINT_T0);
            __mic_su3_gmultdag_vec(u,w);

            _my_pf(&p_yo[cur_offs.isite_upz].y[3], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[4], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_upz].y[5], _MM_HINT_T0);
            __mic_mult_proj_post_back_t(yy,w);

          }
#endif

#ifdef _HOP_Z
          ////////////////////////////////////////
          // Z-forward
          ////////////////////////////////////////
          {

            register __mic_half_spinor_t  w1,w2,w3;

#if _NDIMZ != 1
            if ( iz < NZ - 1 ) {
#else
            {
#endif
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upz];
              _my_pf(&p_yo[ pf_offs.isite_upz].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upz].y[5], _MM_HINT_T1);
              _my_pf(&r_ue.v[2].u[0], _MM_HINT_T0);
              _my_pf(&r_ue.v[2].u[1], _MM_HINT_T0);
              _my_pf(&r_ue.v[2].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_forw_z(w,yptr);

#if _NDIMZ != 1
            } else { // load from receive buffer

//              int ibuff = COL*(itht + NTHT*(iy + NY*ix));
//              w1 = _my_load_ps(&(rbuff_z_up[ibuff+0]));
//              w2 = _my_load_ps(&(rbuff_z_up[ibuff+1]));
//              w3 = _my_load_ps(&(rbuff_z_up[ibuff+2]));
              w1 = _my_load_ps(&(rbuff_z_up[cur_offs.ibuff_z+0]));
              w2 = _my_load_ps(&(rbuff_z_up[cur_offs.ibuff_z+1]));
              w3 = _my_load_ps(&(rbuff_z_up[cur_offs.ibuff_z+2]));
              _my_pf(&(rbuff_z_up[pf_offs.ibuff_z+0]), _MM_HINT_T1);
              _my_pf(&(rbuff_z_up[pf_offs.ibuff_z+1]), _MM_HINT_T1);
              _my_pf(&(rbuff_z_up[pf_offs.ibuff_z+2]), _MM_HINT_T1);

            }
#else
            }
#endif

            {
              register __mic_half_spinor_t  u1,u2,u3;
              const mic_su3mat *uptr = &r_ue.v[2];
              _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[0], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[1], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[2], _MM_HINT_T2);
              __mic_su3_load_u(u,uptr);

              _my_pf(&p_yo[cur_offs.isite_dnz].y[0], _MM_HINT_T0);
              _my_pf(&p_yo[cur_offs.isite_dnz].y[1], _MM_HINT_T0);
              _my_pf(&p_yo[cur_offs.isite_dnz].y[2], _MM_HINT_T0);
              __mic_su3_gmult_vec(u,w);
            }

            _my_pf(&p_yo[cur_offs.isite_dnz].y[3], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_dnz].y[4], _MM_HINT_T0);
            _my_pf(&p_yo[cur_offs.isite_dnz].y[5], _MM_HINT_T0);
            __mic_mult_proj_post_forw_z(yy,w);

          }

          ////////////////////////////////////////
          // Z-backward
          ////////////////////////////////////////
          {

            register __mic_half_spinor_t  w1,w2,w3;
#if _NDIMZ != 1
            if ( 0 < iz ) {
#else
            {
#endif
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnz];
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnz].y[5], _MM_HINT_T1);
              _my_pf(&p_uo[cur_offs.isite_dnz].v[2].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnz].v[2].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnz].v[2].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_z(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnz].v[2];
                _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[0], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[1], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[2], _MM_HINT_T2);
                __mic_su3_load_u(u,uptr);

                __mic_su3_gmultdag_vec(u,w);
              }

#if _NDIMZ != 1
            } else {

//              int ibuff = COL*(itht + NTHT*(iy + NY*ix));
//              w1 = _my_load_ps(&(rbuff_z_dn[ibuff+0]));
//              w2 = _my_load_ps(&(rbuff_z_dn[ibuff+1]));
//              w3 = _my_load_ps(&(rbuff_z_dn[ibuff+2]));
              w1 = _my_load_ps(&(rbuff_z_dn[cur_offs.ibuff_z+0]));
              w2 = _my_load_ps(&(rbuff_z_dn[cur_offs.ibuff_z+1]));
              w3 = _my_load_ps(&(rbuff_z_dn[cur_offs.ibuff_z+2]));
              _my_pf(&(rbuff_z_dn[pf_offs.ibuff_z+0]), _MM_HINT_T1);
              _my_pf(&(rbuff_z_dn[pf_offs.ibuff_z+1]), _MM_HINT_T1);
              _my_pf(&(rbuff_z_dn[pf_offs.ibuff_z+2]), _MM_HINT_T1);

            }
#else
            }
#endif
            __mic_mult_proj_post_back_z(yy,w);

          }
#endif

#ifdef _HOP_Y
          ////////////////////////////////////////
          // Y-forward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;

#if _NDIMY != 1
            if ( iy < NY - 1 ) {
#else
            {
#endif
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upy];
              _my_pf(&p_yo[ pf_offs.isite_upy].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upy].y[5], _MM_HINT_T1);
              _my_pf(&r_ue.v[1].u[0], _MM_HINT_T0);
              _my_pf(&r_ue.v[1].u[1], _MM_HINT_T0);
              _my_pf(&r_ue.v[1].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_forw_y(w,yptr);

#if _NDIMY != 1
            } else { // load from receive buffer

//              int ibuff = COL*(itht + NTHT*(iz + NZ*ix));
//              w1 = _my_load_ps(&(rbuff_y_up[ibuff+0]));
//              w2 = _my_load_ps(&(rbuff_y_up[ibuff+1]));
//              w3 = _my_load_ps(&(rbuff_y_up[ibuff+2]));
              w1 = _my_load_ps(&(rbuff_y_up[cur_offs.ibuff_y+0]));
              w2 = _my_load_ps(&(rbuff_y_up[cur_offs.ibuff_y+1]));
              w3 = _my_load_ps(&(rbuff_y_up[cur_offs.ibuff_y+2]));
              _my_pf(&(rbuff_y_up[pf_offs.ibuff_y+0]), _MM_HINT_T1);
              _my_pf(&(rbuff_y_up[pf_offs.ibuff_y+1]), _MM_HINT_T1);
              _my_pf(&(rbuff_y_up[pf_offs.ibuff_y+2]), _MM_HINT_T1);

            }
#else
            }
#endif

            {
              const mic_su3mat *uptr = &r_ue.v[1];
              _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[0], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[1], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[2], _MM_HINT_T2);

              register __mic_half_spinor_t  u1,u2,u3;
              __mic_su3_load_u(u,uptr);

              __mic_su3_gmult_vec(u,w);
            }

            __mic_mult_proj_post_forw_y(yy,w);

          }

          ////////////////////////////////////////
          // Y-backward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;
#if _NDIMY != 1
            if ( 0 < iy ) {
#else
            {
#endif
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dny];
              _my_pf(&p_yo[ pf_offs.isite_dny].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dny].y[5], _MM_HINT_T1);
              _my_pf(&p_uo[cur_offs.isite_dny].v[1].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dny].v[1].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dny].v[1].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_y(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_dny].v[1];
                _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[0], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[1], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[2], _MM_HINT_T2);
                __mic_su3_load_u(u,uptr);

                __mic_su3_gmultdag_vec(u,w);
              }

#if _NDIMY != 1
            } else {

//              int ibuff = COL*(itht + NTHT*(iz + NZ*ix));
//              w1 = _my_load_ps(&(rbuff_y_dn[ibuff+0]));
//              w2 = _my_load_ps(&(rbuff_y_dn[ibuff+1]));
//              w3 = _my_load_ps(&(rbuff_y_dn[ibuff+2]));
              w1 = _my_load_ps(&(rbuff_y_dn[cur_offs.ibuff_y+0]));
              w2 = _my_load_ps(&(rbuff_y_dn[cur_offs.ibuff_y+1]));
              w3 = _my_load_ps(&(rbuff_y_dn[cur_offs.ibuff_y+2]));
              _my_pf(&(rbuff_y_dn[pf_offs.ibuff_y+0]), _MM_HINT_T1);
              _my_pf(&(rbuff_y_dn[pf_offs.ibuff_y+1]), _MM_HINT_T1);
              _my_pf(&(rbuff_y_dn[pf_offs.ibuff_y+2]), _MM_HINT_T1);

            }
#else
            }
#endif
            __mic_mult_proj_post_back_y(yy,w);

          }
#endif

#ifdef _HOP_X
          ////////////////////////////////////////
          // X-forward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;

#if _NDIMX != 1
            if ( ix < NX - 1 ) {
#else
            {
#endif
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upx];
              _my_pf(&p_yo[ pf_offs.isite_upx].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_upx].y[5], _MM_HINT_T1);
              _my_pf(&r_ue.v[0].u[0], _MM_HINT_T0);
              _my_pf(&r_ue.v[0].u[1], _MM_HINT_T0);
              _my_pf(&r_ue.v[0].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_forw_x(w,yptr);

#if _NDIMX != 1
            } else { // load from receive buffer

//              int ibuff = COL*(itht + NTHT*(iy + NY*iz));
//              w1 = _my_load_ps(&(rbuff_x_up[ibuff+0]));
//              w2 = _my_load_ps(&(rbuff_x_up[ibuff+1]));
//              w3 = _my_load_ps(&(rbuff_x_up[ibuff+2]));
              w1 = _my_load_ps(&(rbuff_x_up[cur_offs.ibuff_x+0]));
              w2 = _my_load_ps(&(rbuff_x_up[cur_offs.ibuff_x+1]));
              w3 = _my_load_ps(&(rbuff_x_up[cur_offs.ibuff_x+2]));
              _my_pf(&(rbuff_x_up[pf_offs.ibuff_x+0]), _MM_HINT_T1);
              _my_pf(&(rbuff_x_up[pf_offs.ibuff_x+1]), _MM_HINT_T1);
              _my_pf(&(rbuff_x_up[pf_offs.ibuff_x+2]), _MM_HINT_T1);

            }
#else
            }
#endif

            {
              register __mic_half_spinor_t  u1,u2,u3;

              const mic_su3mat *uptr = &r_ue.v[0];
              _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[0], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[1], _MM_HINT_T2);
              _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[2], _MM_HINT_T2);
              __mic_su3_load_u(u,uptr);

              __mic_su3_gmult_vec(u,w);
            }

            __mic_mult_proj_post_forw_x(yy,w);

          }

          ////////////////////////////////////////
          // X-backward
          ////////////////////////////////////////
          {
            register __mic_half_spinor_t  w1,w2,w3;
#if _NDIMX != 1
            if ( 0 < ix ) {
#else
            {
#endif
              const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnx];
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[0], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[1], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[2], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[3], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[4], _MM_HINT_T1);
              _my_pf(&p_yo[ pf_offs.isite_dnx].y[5], _MM_HINT_T1);
              _my_pf(&p_uo[cur_offs.isite_dnx].v[0].u[0], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnx].v[0].u[1], _MM_HINT_T0);
              _my_pf(&p_uo[cur_offs.isite_dnx].v[0].u[2], _MM_HINT_T0);
              __mic_mult_proj_pre_back_x(w,yptr);

              {
                register __mic_half_spinor_t  u1,u2,u3;
                const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnx].v[0];
                _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[0], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[1], _MM_HINT_T2);
                _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[2], _MM_HINT_T2);
                __mic_su3_load_u(u,uptr);

                __mic_su3_gmultdag_vec(u,w);
              }

#if _NDIMX != 1
            } else {

//              int ibuff = COL*(itht + NTHT*(iy + NY*iz));
//              w1 = _my_load_ps(&(rbuff_x_dn[ibuff+0]));
//              w2 = _my_load_ps(&(rbuff_x_dn[ibuff+1]));
//              w3 = _my_load_ps(&(rbuff_x_dn[ibuff+2]));
              w1 = _my_load_ps(&(rbuff_x_dn[cur_offs.ibuff_x+0]));
              w2 = _my_load_ps(&(rbuff_x_dn[cur_offs.ibuff_x+1]));
              w3 = _my_load_ps(&(rbuff_x_dn[cur_offs.ibuff_x+2]));
              _my_pf(&(rbuff_x_dn[pf_offs.ibuff_x+0]), _MM_HINT_T1);
              _my_pf(&(rbuff_x_dn[pf_offs.ibuff_x+1]), _MM_HINT_T1);
              _my_pf(&(rbuff_x_dn[pf_offs.ibuff_x+2]), _MM_HINT_T1);

            }
#else
            }
#endif
            __mic_mult_proj_post_back_x(yy,w);

          }
#endif

#ifdef _HOP_CLV
#ifdef _CLOVER
          {
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[0], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[1], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[2], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[3], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[4], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[5], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[6], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[7], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[8], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[0], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[1], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[2], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[3], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[4], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[5], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[6], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[7], _MM_HINT_T2);
            _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[8], _MM_HINT_T2);
            __mic_mult_clvoer_term(yy,&fptr[cur_offs.isite_ctr]);
          }
#endif
#endif

          {
            _my_ngostore_ps(&((*mye).s[cur_offs.isite_ctr].y[0]),yy1);
            _my_ngostore_ps(&((*mye).s[cur_offs.isite_ctr].y[1]),yy2);
            _my_ngostore_ps(&((*mye).s[cur_offs.isite_ctr].y[2]),yy3);
            _my_ngostore_ps(&((*mye).s[cur_offs.isite_ctr].y[3]),yy4);
            _my_ngostore_ps(&((*mye).s[cur_offs.isite_ctr].y[4]),yy5);
            _my_ngostore_ps(&((*mye).s[cur_offs.isite_ctr].y[5]),yy6);
          }

        }

      }  // end of while
    }    // end of omp parallel

  }
  stopTimer(solver_timer.mult_pst_timer);
  stopTimer(solver_timer.mult_timer);
#endif

#ifdef _TIMING_
#ifndef _singlePU
  solver_timer.ycopy_count++;
  solver_timer.ycopy_dsize += 2*(NRAS_GHOST_Z+NRAS_GHOST_Y+NRAS_GHOST_X)*sizeof(__mic_half_spinor_t);
#endif
#endif
  solver_timer.mult_count++;

#ifdef _TIMING_

  solver_timer.mult_flop_count += (MULT_HOP+MULT_CLOVER)*NTH*NZ*NY*NX;

  int mult_pre_flop_count = 0;
  int mult_in_flop_count  = 0;
  int mult_pst_flop_count = 0;

#if _NDIMZ != 1
  mult_pre_flop_count += (12       )*NTH*NY*NX   // Z-forw
                        +(12 + 66*2)*NTH*NY*NX;  // Z-back
#endif
#if _NDIMY != 1
  mult_pre_flop_count += (12       )*NTH*NZ*NX   // Y-forw
                        +(12 + 66*2)*NTH*NZ*NX;  // Y-back
#endif
#if _NDIMX != 1
  mult_pre_flop_count += (12       )*NTH*NZ*NY   // X-forw
                        +(12 + 66*2)*NTH*NZ*NY;  // X-back
#endif

#if   _PAR_DIR == 1
  mult_in_flop_count  += (MULT_HOP+MULT_CLOVER)*NTH*(NZ-2)*NY*NX;
#elif _PAR_DIR == 2
  mult_in_flop_count  += (MULT_HOP+MULT_CLOVER)*NTH*NZ*(NY-2)*NX;
#elif _PAR_DIR == 3
  mult_in_flop_count  += (MULT_HOP+MULT_CLOVER)*NTH*(NZ-2)*(NY-2)*NX;
#elif _PAR_DIR == 4
  mult_in_flop_count  += (MULT_HOP+MULT_CLOVER)*NTH*NZ*NY*(NX-2);
#elif _PAR_DIR == 5
  mult_in_flop_count  += (MULT_HOP+MULT_CLOVER)*NTH*(NZ-2)*NY*(NX-2);
#elif _PAR_DIR == 6
  mult_in_flop_count  += (MULT_HOP+MULT_CLOVER)*NTH*NZ*(NY-2)*(NX-2);
#elif _PAR_DIR == 7
  mult_in_flop_count  += (MULT_HOP+MULT_CLOVER)*NTH*(NZ-2)*(NY-2)*(NX-2);
#endif

  mult_pst_flop_count += (MULT_HOP+MULT_CLOVER)*NTH*NZ*NY*NX - mult_in_flop_count - mult_pre_flop_count;

  solver_timer.mult_pre_flop_count += mult_pre_flop_count;
  solver_timer.mult_in_flop_count  += mult_in_flop_count;
  solver_timer.mult_pst_flop_count += mult_pst_flop_count;

#endif

  stopTimer(solver_timer.mult_w_comm_timer);

}

#undef _HOP_T
#undef _HOP_Z
#undef _HOP_Y
#undef _HOP_X
#undef _HOP_CLV

#endif
