#ifndef _ASSIGN_MULT_HOP_EO_RAS_MIC_H
#define _ASSIGN_MULT_HOP_EO_RAS_MIC_H

#define _HOP_T
#define _HOP_Z
#define _HOP_Y
#define _HOP_X
#define _HOP_CLV


extern "C" void
#ifdef _CLOVER
assign_mult_hop_eo_ras_mic(      mic_ras_wqf_eo   *restrict mye, 
                           const mic_ras_wqf_eo   *restrict  yo,
                           const mic_ras_gvf_eo_soa   *restrict  ue,
                           const mic_ras_gvf_eo_soa   *restrict  uo,
                           const mic_ras_clvf_eo       *fclinve) 
#else
assign_mult_hop_eo_ras_mic(      mic_ras_wqf_eo   *restrict mye, 
                           const mic_ras_wqf_eo   *restrict  yo,
                           const mic_ras_gvf_eo_soa   *restrict  ue,
                           const mic_ras_gvf_eo_soa   *restrict  uo)
#endif
////////////////////////////////////////////////////////////////////////////////
// Multiply hopping matrix on the even/odd site fermion field
// with extended region.
//
//  mye = Fee Meo yo
//
//  mye : even/odd site fermion field with extended region (output)
//   yo : odd/even site fermion field with extended region (input)
//   ue : even/odd site gauge field with extended region (input)
//   uo : odd/even site gauge field with extended region (input)
//  fclinve : even/odd site inverse clover term with extented resion (input)
//
////////////////////////////////////////////////////////////////////////////////
{

//printf("RasMult: mye %p yo %p ue %p uo %p fclinve %p\n",
//mye->s,yo->s,ue->s,uo->s,fclinve->s);
  __assume_aligned((void *)mye, _ALIGN_SIZE);
  __assume_aligned((void *) yo, _ALIGN_SIZE);
  __assume_aligned((void *) ue, _ALIGN_SIZE);
  __assume_aligned((void *) uo, _ALIGN_SIZE);
#ifdef _CLOVER
  __assume_aligned((void *)fclinve, _ALIGN_SIZE);
#endif

  startTimer(solver_timer.ras_mult_timer);

  int ipeo = mic_ipeo;
  int ieo  = (*ue).ieo;

  (*mye).ieo  = ieo;

  //
  // site loop for extended domain
  //
  const int tilex = 2, tiley = 2, tilez = 4, tilet = 8;
  Percore cores[maxCores];
  const int nblocksx = (NEX + tilex - 1) / tilex;
  const int nblocksy = (NEY + tiley - 1) / tiley;
  const int nblocksz = (NEZ + tilez - 1) / tilez;
  const int nblockst = (NTHT + tilet - 1) / tilet;
  const int nblocks = nblocksx * nblocksy * nblocksz * nblockst;
  const int offsize = nblocks * tilex * tiley * tilez * tilet;
  static struct offs {
    int isite_ctr;
    int isite_upt, isite_dnt;
    int isite_upz, isite_dnz;
    int isite_upy, isite_dny;
    int isite_upx, isite_dnx;
  } offsets[offsize+1];
  static bool offsinit = false;
  if (!offsinit) {
    offsinit = true;
    #pragma omp parallel
  {
    int nblocks = nblocksx * nblocksy * nblocksz * nblockst;
    Sched sch(nblocks, cores);
    #pragma omp barrier
    int block;
    while ((block = sch.nextiter()) != -1)
    {
      int iix = block / (nblocksy*nblocksz*nblockst);
      int rem = block % (nblocksy*nblocksz*nblockst);
      int iiy = rem   / (         nblocksz*nblockst);
      rem     = rem   % (         nblocksz*nblockst);
      int iiz = rem   / (                  nblockst);
      int iit = rem   % (                  nblockst);

      int ioffs = (iit + nblockst*(iiz + nblocksz*(iiy + nblocksy*iix)));
      ioffs *= tilex * tiley * tilez * tilet;

      iiz *= tilez;
      iiy *= tiley;
      iix *= tilex;
      iit *= tilet;
      int zmax = iiz + tilez; if (zmax > NEZ) zmax = NEZ;
      int ymax = iiy + tiley; if (ymax > NEY) ymax = NEY;
      int xmax = iix + tilex; if (xmax > NEX) xmax = NEX;
      int tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;

      for (int ix = iix; ix < xmax; ++ix)
      for (int iy = iiy; iy < ymax; ++iy)
      for (int iz = iiz; iz < zmax; ++iz)
      for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
          int ioffs1 = ioffs + sch.myht;
          if (ioffs1 >= offsize)
          {
            fprintf(stderr, "Error: overflow: ioffs %d >= %d\n",
              ioffs, offsize);
            exit(1);
          }
          offsets[ioffs1].isite_ctr = get_ras_site(itht, iz, iy, ix);
          //printf("%d: %d\n", ioffs, offsets[ioffs].isite_ctr);
          offsets[ioffs1].isite_upt = get_ras_site((itht+1)%NTHT,iz,iy,ix);
          offsets[ioffs1].isite_dnt = get_ras_site((itht-1+NTHT)%NTHT,iz,iy,ix);
          offsets[ioffs1].isite_upz = get_ras_site(itht,(iz+1)%NEZ,iy,ix);
          offsets[ioffs1].isite_dnz = get_ras_site(itht,(iz-1+NEZ)%NEZ,iy,ix);
          offsets[ioffs1].isite_upy = get_ras_site(itht,iz,(iy+1)%NEY,ix);
          offsets[ioffs1].isite_dny = get_ras_site(itht,iz,(iy-1+NEY)%NEY,ix);
          offsets[ioffs1].isite_upx = get_ras_site(itht,iz,iy,(ix+1)%NEX);
          offsets[ioffs1].isite_dnx = get_ras_site(itht,iz,iy,(ix-1+NEX)%NEX);
          ioffs += nHT;
      }
    }
  } // parallel
  offsets[offsize] = offsets[offsize-1];
  }


  #ifdef PMU_STAT
  solver_timer.ras_mult_stat.start();
  #endif
#pragma omp parallel
{
  Sched sch(nblocks, cores);
  #pragma omp barrier
  #ifdef PMU_STAT
  solver_timer.ras_mult_stat.enter(sch.mythread);
  #endif
  int block;
  while ((block = sch.nextiter()) != -1)
  {
    int iix = block / (nblocksy*nblocksz*nblockst);
    int rem = block % (nblocksy*nblocksz*nblockst);
    int iiy = rem   / (         nblocksz*nblockst);
    rem     = rem   % (         nblocksz*nblockst);
    int iiz = rem   / (                  nblockst);
    int iit = rem   % (                  nblockst);

    int ioffs = (iit + nblockst*(iiz + nblocksz*(iiy + nblocksy*iix)));
    ioffs *= tilex * tiley * tilez * tilet;

    iiz *= tilez;
    iiy *= tiley;
    iix *= tilex;
    iit *= tilet;
    int zmax = iiz + tilez; if (zmax > NEZ) zmax = NEZ;
    int ymax = iiy + tiley; if (ymax > NEY) ymax = NEY;
    int xmax = iix + tilex; if (xmax > NEX) xmax = NEX;
    int tmax = iit + tilet; if (tmax > NTHT) tmax = NTHT;


    for (int ix = iix; ix < xmax; ++ix)
    for (int iy = iiy; iy < ymax; ++iy)
    for (int iz = iiz; iz < zmax; ++iz)
    for (int itht = iit+sch.myht; itht < tmax; itht+=nHT) {
      int ieoxyz = (ipeo + (ix-NDEPTHX+1) + (iy-NDEPTHY+1) + (iz-NDEPTHZ+1) + ieo + 256) % 2; // add 256 to ensure positive ieoxyz
      int ioffs1 = ioffs + sch.myht;
      const offs &cur_offs = offsets[ioffs1];
      const offs &pf_offs = offsets[ioffs1+nHT];
      ioffs += nHT;
      const mic_ras_gsf_eo_soa *p_ue = (*ue).s;
      const mic_ras_gsf_eo_soa &r_ue = p_ue[cur_offs.isite_ctr];
      const mic_ras_gsf_eo_soa *p_uo = (*uo).s;
      const mic_su3_spinor *p_yo = (*yo).s;
#ifdef _CLOVER
      const mic_clover *fptr = (*fclinve).s;
#endif

      register __mic_half_spinor_t yy1,yy2,yy3,yy4,yy5,yy6;
      { 
        register __mic_half_spinor_t zero = _my_setzero_ps();

        yy1 = zero;
        yy2 = zero;
        yy3 = zero;
        yy4 = zero;
        yy5 = zero;
        yy6 = zero;
      }

#ifdef _HOP_T
      ////////////////////////////////////////
      // T-forward
      ////////////////////////////////////////
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
        __mic_mult_proj_pre_forw_t(w,yptr);

        if ( 0 == ieoxyz ) {

          register __mic_half_spinor_t  s1,s2,s3;

          // periodic boundary condition in T-direction

          const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upt];
          __mic_mult_proj_pre_forw_t(s,yptr);

          __mic_mult_shift_forw_t(w,s);

        }

        register __mic_half_spinor_t  u1,u2,u3;
        const mic_su3mat *uptr = &r_ue.v[3];
        _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[0], _MM_HINT_T2);
        _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[1], _MM_HINT_T2);
        _my_pf(&p_ue[pf_offs.isite_ctr].v[3].u[2], _MM_HINT_T2);
        __mic_su3_load_u(u,uptr);

        __mic_su3_gmult_vec(u,w);

        __mic_mult_proj_post_forw_t(yy,w);

      }

      ////////////////////////////////////////
      // T-backward
      ////////////////////////////////////////
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_ctr];
        __mic_mult_proj_pre_back_t(w,yptr);

        register __mic_half_spinor_t  u1,u2,u3;
        const mic_su3mat *uptr = &p_uo[cur_offs.isite_ctr].v[3];
        _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[0], _MM_HINT_T1);
        _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[1], _MM_HINT_T1);
        _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[2], _MM_HINT_T1);
        __mic_su3_load_u(u,uptr);

        if ( 1 == ieoxyz ) {

          // periodic boundary condition in T-direction
          register __mic_half_spinor_t  s1,s2,s3;

          const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnt];
          __mic_mult_proj_pre_back_t(s,yptr);

          __mic_mult_shift_back_t(w,s);

          const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnt].v[3];
          _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[0], _MM_HINT_T1);
          _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[1], _MM_HINT_T1);
          _my_pf(&p_uo[pf_offs.isite_ctr].v[3].u[2], _MM_HINT_T1);
          __mic_su3_load_u(s,uptr);

          __mic_mult_shift_back_t(u,s);

        }

        __mic_su3_gmultdag_vec(u,w);

        __mic_mult_proj_post_back_t(yy,w);

      }
#endif

#ifdef _HOP_Z
      ////////////////////////////////////////
      // Z-forward
      ////////////////////////////////////////
#if _NDIMZ != 1
      if ( iz < NEZ - 1 )
#endif
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upz];
        _my_pf(&p_yo[pf_offs.isite_upz].y[0], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upz].y[1], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upz].y[2], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upz].y[3], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upz].y[4], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upz].y[5], _MM_HINT_T1);
        __mic_mult_proj_pre_forw_z(w,yptr);

        {
          register __mic_half_spinor_t  u1,u2,u3;
          const mic_su3mat *uptr = &r_ue.v[2];
          _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[0], _MM_HINT_T2);
          _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[1], _MM_HINT_T2);
          _my_pf(&p_ue[pf_offs.isite_ctr].v[2].u[2], _MM_HINT_T2);
          __mic_su3_load_u(u,uptr);

          __mic_su3_gmult_vec(u,w);
        }

        __mic_mult_proj_post_forw_z(yy,w);

      }

      ////////////////////////////////////////
      // Z-backward
      ////////////////////////////////////////
#if _NDIMZ != 1
      if ( 0 < iz )
#endif
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnz];
        _my_pf(&p_yo[pf_offs.isite_dnz].y[0], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnz].y[1], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnz].y[2], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnz].y[3], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnz].y[4], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnz].y[5], _MM_HINT_T1);
        __mic_mult_proj_pre_back_z(w,yptr);

        {
          register __mic_half_spinor_t  u1,u2,u3;
          const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnz].v[2];
          _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[0], _MM_HINT_T2);
          _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[1], _MM_HINT_T2);
          _my_pf(&p_uo[pf_offs.isite_dnz].v[2].u[2], _MM_HINT_T2);
          __mic_su3_load_u(u,uptr);

          __mic_su3_gmultdag_vec(u,w);
        }

        __mic_mult_proj_post_back_z(yy,w);

      }
#endif

#ifdef _HOP_Y
      ////////////////////////////////////////
      // Y-forward
      ////////////////////////////////////////
#if _NDIMY != 1
      if ( iy < NEY - 1 )
#endif
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upy];
        _my_pf(&p_yo[pf_offs.isite_upy].y[0], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upy].y[1], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upy].y[2], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upy].y[3], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upy].y[4], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upy].y[5], _MM_HINT_T1);
        __mic_mult_proj_pre_forw_y(w,yptr);

        {
          register __mic_half_spinor_t  u1,u2,u3;
          const mic_su3mat *uptr = &r_ue.v[1];
          _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[0], _MM_HINT_T2);
          _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[1], _MM_HINT_T2);
          _my_pf(&p_ue[pf_offs.isite_ctr].v[1].u[2], _MM_HINT_T2);
          __mic_su3_load_u(u,uptr);

          __mic_su3_gmult_vec(u,w);
        }

        __mic_mult_proj_post_forw_y(yy,w);

      }

      ////////////////////////////////////////
      // Y-backward
      ////////////////////////////////////////
#if _NDIMY != 1
      if ( 0 < iy )
#endif
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dny];
        _my_pf(&p_yo[pf_offs.isite_dny].y[0], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dny].y[1], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dny].y[2], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dny].y[3], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dny].y[4], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dny].y[5], _MM_HINT_T1);
        __mic_mult_proj_pre_back_y(w,yptr);

        {
          register __mic_half_spinor_t  u1,u2,u3;
          const mic_su3mat *uptr = &p_uo[cur_offs.isite_dny].v[1];
          _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[0], _MM_HINT_T2);
          _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[1], _MM_HINT_T2);
          _my_pf(&p_uo[pf_offs.isite_dny].v[1].u[2], _MM_HINT_T2);
          __mic_su3_load_u(u,uptr);

          __mic_su3_gmultdag_vec(u,w);
        }

        __mic_mult_proj_post_back_y(yy,w);

      }
#endif

#ifdef _HOP_X
      ////////////////////////////////////////
      // X-forward
      ////////////////////////////////////////
#if _NDIMX != 1
      if ( ix < NEX - 1 )
#endif
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_upx];
        _my_pf(&p_yo[pf_offs.isite_upx].y[0], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upx].y[1], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upx].y[2], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upx].y[3], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upx].y[4], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_upx].y[5], _MM_HINT_T1);
        __mic_mult_proj_pre_forw_x(w,yptr);

        {
          register __mic_half_spinor_t  u1,u2,u3;
          const mic_su3mat *uptr = &r_ue.v[0];
          _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[0], _MM_HINT_T2);
          _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[1], _MM_HINT_T2);
          _my_pf(&p_ue[pf_offs.isite_ctr].v[0].u[2], _MM_HINT_T2);
          __mic_su3_load_u(u,uptr);

          __mic_su3_gmult_vec(u,w);
        }

        __mic_mult_proj_post_forw_x(yy,w);

      }

      ////////////////////////////////////////
      // X-backward
      ////////////////////////////////////////
#if _NDIMX != 1
      if ( 0 < ix )
#endif
      {
        register __mic_half_spinor_t  w1,w2,w3;
        const mic_su3_spinor *yptr = &p_yo[cur_offs.isite_dnx];
        _my_pf(&p_yo[pf_offs.isite_dnx].y[0], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnx].y[1], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnx].y[2], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnx].y[3], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnx].y[4], _MM_HINT_T1);
        _my_pf(&p_yo[pf_offs.isite_dnx].y[5], _MM_HINT_T1);
        __mic_mult_proj_pre_back_x(w,yptr);

        {
          register __mic_half_spinor_t  u1,u2,u3;
          const mic_su3mat *uptr = &p_uo[cur_offs.isite_dnx].v[0];
          _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[0], _MM_HINT_T2);
          _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[1], _MM_HINT_T2);
          _my_pf(&p_uo[pf_offs.isite_dnx].v[0].u[2], _MM_HINT_T2);
          __mic_su3_load_u(u,uptr);

          __mic_su3_gmultdag_vec(u,w);
        }

        __mic_mult_proj_post_back_x(yy,w);

      }
#endif

#ifdef _HOP_CLV
#ifdef _CLOVER
      {
        #if 1
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[0], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[1], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[2], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[3], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[4], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[5], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[6], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[7], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[0].f[8], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[0], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[1], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[2], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[3], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[4], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[5], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[6], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[7], _MM_HINT_T2);
        _my_pf(&fptr[pf_offs.isite_ctr].c[1].f[8], _MM_HINT_T2);
        #endif
        __mic_mult_clvoer_term(yy,&fptr[cur_offs.isite_ctr]);
      }
#endif
#endif

      {
        __mic_half_spinor_t *ye = &(*mye).s[cur_offs.isite_ctr].y[0];
        _my_ngostore_ps(&ye[0],yy1);
        _my_ngostore_ps(&ye[1],yy2);
        _my_ngostore_ps(&ye[2],yy3);
        _my_ngostore_ps(&ye[3],yy4);
        _my_ngostore_ps(&ye[4],yy5);
        _my_ngostore_ps(&ye[5],yy6);
      }


    }
  }
  #ifdef PMU_STAT
  solver_timer.ras_mult_stat.exit(sch.mythread);
  #endif
}
  #ifdef PMU_STAT
  solver_timer.ras_mult_stat.accum(omp_get_max_threads());
  #endif

  solver_timer.ras_mult_count++;
  stopTimer(solver_timer.ras_mult_timer);
#ifdef _TIMING_

#ifdef _CLOVER
  solver_timer.ras_mult_flop_count += (MULT_HOP+MULT_CLOVER)*NTH*NEZ*NEY*NEX;
#else
  solver_timer.ras_mult_flop_count += (MULT_HOP            )*NTH*NEZ*NEY*NEX;
#endif

#if _NDIMZ != 1
  solver_timer.ras_mult_flop_count += -2*168*NTH*NEZ*NEX;
#endif
#if _NDIMY != 1
  solver_timer.ras_mult_flop_count += -2*168*NTH*NEY*NEX;
#endif
#if _NDIMX != 1
  solver_timer.ras_mult_flop_count += -2*168*NTH*NEZ*NEY;
#endif

#endif

}

#undef _HOP_T
#undef _HOP_Z
#undef _HOP_Y
#undef _HOP_X
#undef _HOP_CLV

#endif
