#include <immintrin.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <omp.h>
#include "cconfig.h"
#include "cmic_config.h"
#include "mic_param_size.h"
// interface routines for offloaded solver
#include <mpi.h>
#include "signal.h"

    
#include "csolver.h"
#pragma offload_attribute(push, target(mic))
#include "scif_interface.h"
#pragma offload_attribute(pop)

SCIF_Interface *scif_host;
__declspec(target(mic)) SCIF_Interface *scif_mic;
static int mic_targetid;

__declspec(target(mic)) static char *mic_affinity;

__declspec(target(mic)) void
segv_handler(int sig)
{
    printf("MIC: segv_handler: %d\n", sig);
    fflush(0);
    _exit(1);
}

void
sigint_handler(int sig)
{
    printf("Host: caught sigint\n");
    fflush(0);
    if (scif_host)
        scif_host->close();
    _exit(1);
}

extern "C" void __declspec(target(mic)) setnodeid(const cmic_solver *s);
extern "C" void __declspec(target(mic)) unset_ras_fields();
extern "C" void __declspec(target(mic)) new_mic_ras_solver();
extern "C" __declspec(target(mic)) void
assign_inv_mult_eoprec_wd_bicgstab_mic(const float  *skappa,
                                       const float  *stol,
                                       int          *iter,
                                             mic_wqf_eo   *restrict yde,
                                       const mic_wqf_eo   *restrict  ye);
extern "C" __declspec(target(mic)) void
set_ras_fields(const mic_gvf_soa     *u,
               const mic_ras_gvf_soa *ru,
               const mic_clvf        *fclinv,
               const mic_ras_clvf    *rfclinv);

static void
parsecards()
{
#ifndef _singlePU
    int rank;
    int nranks;
    MPI_Comm_size(MPI_COMM_WORLD, &nranks);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    int mics[nranks];
    char *affinities[nranks];
    char *p = getenv("CCS_CARDS");
    if (p == 0)
        for (int i = 0; i < nranks; ++i)
            mics[i] = 0;
    else
    {
        char cards[strlen(p)+1];
        strcpy(cards, p);
        int i = 0;
        char *q;
        for (q = strtok(cards, ","); q != 0; q = strtok(0, ","))
        {
            if (i >= nranks)
                break;
            int card = atoi(q);
            mics[i++] = card;
        }
        int lastrank = i;
        i = 0;
        for (int j = lastrank; j < nranks; ++j)
        {
            mics[j] = mics[i];
            ++i;
            if (i == lastrank)
                i = 0;
        }
    }
    mic_targetid = mics[rank];
    p = getenv("CCS_AFFINITIES");
    if (p == 0)
        for (int i = 0; i < nranks; ++i)
            affinities[i] = 0;
    else
    {
        p = strdup(p);
        int i = 0;
        char *q;
        for (q = strtok(p, ","); q != 0; q = strtok(0, ","))
        {
            if (i >= nranks)
                break;
            printf("q = %s\n", q);
            affinities[i++] = q;
        }
        int lastrank = i;
        i = 0;
        for (int j = lastrank; j < nranks; ++j)
        {
            affinities[j] = affinities[i];
            ++i;
            if (i == lastrank)
                i = 0;
        }
    }
    mic_targetid = mics[rank];
    mic_affinity = affinities[rank];
    printf("%d: MIC TARGET %d affinity %s\n", rank, mic_targetid, mic_affinity);
#endif
}


extern "C" void
start_scif()
{
    int offload_signal = 0xF;
    int mic_node;
    uint16_t mic_port;
    parsecards();
    if (mic_affinity == 0)
        mic_affinity = "";
    #pragma offload target(mic:mic_targetid) nocopy(scif_mic) \
        in(mic_affinity:length(strlen(mic_affinity)+1)) \
        out(mic_node) out(mic_port)
    {
        signal(SIGSEGV, segv_handler);
        char aff[1024];
        if (*mic_affinity != 0)
        {
            snprintf(aff, sizeof(aff), "KMP_PLACE_THREADS=%s", mic_affinity);
            kmp_set_defaults(aff);
        }
        printf("MIC creating server port %s\n", aff);
        scif_mic = new SCIF_Interface();
        mic_port = scif_mic->create_receive_port(mic_node);
        printf("MIC created server port\n");
    }
    printf("Host: mic_node = %d mic_port = %d\n", mic_node, mic_port);
    // Start MIC server
    #pragma offload target(mic:mic_targetid) nocopy(scif_mic) \
        signal(offload_signal)
    {
        printf("MIC waiting for connection...\n"); fflush(0);
        scif_mic->server();
        printf("MIC connect successful\n");
    }
    scif_host = new SCIF_Interface();
    scif_host->create_send_port(mic_node, mic_port);
    printf("CPU connected to MIC.\n");
    signal(SIGINT, sigint_handler);
    #pragma offload_wait target(mic:mic_targetid) wait(offload_signal)
}

extern "C" void
new_mic_solver(
    void **fsolver,
    nodeinfo *nodeinfo,
    const double *kappa,
    const int *ieo,
    const mic_gvf_soa *u,
    const mic_ras_gvf_soa *u_ras,
    const mic_clvf *fclinv,
    const mic_ras_clvf *fclinv_ras)
{
    cmic_solver *s;
    float kappa1 = *kappa;
    int ieo1 = *ieo;
    struct nodeinfo ni = *nodeinfo;
    // FIXME: Need to figure out which card to use
    mic_gvf_soa *u1;
    mic_ras_gvf_soa *u_ras1;
    mic_clvf *fclinv1;
    mic_ras_clvf *fclinv_ras1;
    printf("before trasnf cmic_solver ptr = %ld\n",s); 
    // Allocate the target struct and its pointers
    #pragma offload_transfer target(mic:mic_targetid) \
      nocopy(s          [0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(u1         [0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(u_ras1     [0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(fclinv1    [0:1] : alloc_if(1) free_if(0) targetptr) \
      nocopy(fclinv_ras1[0:1] : alloc_if(1) free_if(0) targetptr)

    printf("after transf cmic_solver ptr = %ld\n",s); 

    *fsolver = (void *)s;
    // Transfer the data to the card
    #pragma offload target(mic:mic_targetid) \
        in(s[0:0] : alloc_if(0) free_if(0) targetptr) \
        in(u         [0:1] : into(u1         [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(u_ras     [0:1] : into(u_ras1     [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(fclinv    [0:1] : into(fclinv1    [0:1]) alloc_if(0) free_if(0) targetptr) \
        in(fclinv_ras[0:1] : into(fclinv_ras1[0:1]) alloc_if(0) free_if(0) targetptr)
    {
        s->kappa = kappa1;
        s->ieo = ieo1;
        s->ni = ni;
        s->u = u1;
        s->u_ras = u_ras1;
        s->fclinv = fclinv1;
        s->fclinv_ras = fclinv_ras1;
        setnodeid(s);
        new_mic_ras_solver();
        printf("MIC cmic_solver ptr = %ld\n",s); 
    }
}
extern "C" __declspec(target(mic)) void
delete_mic_ras_solver();

extern "C" void
delete_mic_solver(const void **fsolver)
{
    cmic_solver *s = (cmic_solver *)*fsolver;

    void *u1, *u_ras1, *fclinv1, *fclinv_ras1;
    mic_gvf_soa     *u;
    mic_ras_gvf_soa *u_ras;
    mic_clvf        *fclinv;
    mic_ras_clvf     *fclinv_ras;
    // Get the device pointers
    #pragma offload target(mic:mic_targetid) \
        in(s[0:0] : alloc_if(0) free_if(0) targetptr)
    {
        u1          = (void *)s->u;
        u_ras1      = (void *)s->u_ras;
        fclinv1     = (void *)s->fclinv;
        fclinv_ras1 = (void *)s->fclinv_ras;
        unset_ras_fields();
        delete_mic_ras_solver();
    }
    u          = (mic_gvf_soa *)u1;
    u_ras      = (mic_ras_gvf_soa *)u_ras1;
    fclinv     = (mic_clvf *)fclinv1;
    fclinv_ras = (mic_ras_clvf *)fclinv_ras1;
    // free the device memory
    #pragma offload_transfer target(mic:mic_targetid) \
        in(s[0:0] : alloc_if(0) free_if(1) targetptr) \
        in(u         [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(u_ras     [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(fclinv    [0:0] : alloc_if(0) free_if(1) targetptr) \
        in(fclinv_ras[0:0] : alloc_if(0) free_if(1) targetptr)
}


#include "commands.h"

static void
process_cmds()
{
    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    Command cmd1[MaxCmd];
    assert((sizeof(Cmddata) % sizeof(Command)) == 0);
    assert(sizeof(cmd1) >= sizeof(Command) + sizeof(Cmddata));
    memset(cmd1, 0, sizeof(cmd1));
    int nreqs = 0;
    int ndirty = 0;
    MPI_Request reqs[MaxReq];
    off_t dirtybuf[MaxReq];
    int dirtylen[MaxReq];
    for (;;)
    {
        uint64_t msgstart;
        size_t len;
        int data;
        scif_host->dequeue(data, msgstart, len, true);
        assert(len <= sizeof(cmd1));
        scif_host->qcopy(cmd1, msgstart, len);
        //printf("%d: Command: %d len %lu\n", rank, cmd1->cmd, len);
        if (cmd1->cmd == Command::CMD_Stop)
        {
            break;
        }
        Cmddata cmddata;
        memcpy(&cmddata, cmd1->data, sizeof(cmddata));
        void *buff1, *buff2;
        MPI_Status status;
        switch (cmd1->cmd)
        {
        case CMD_Isend:
            scif_host->readbuf(cmddata.sbuff, cmddata.buflen);
            scif_host->fence();
            buff1 = scif_host->getbuf(cmddata.sbuff);
            assert(nreqs < MaxReq);
            MPI_Isend(buff1, cmddata.buflen, MPI_BYTE, cmddata.snode, 0,
                MPI_COMM_WORLD, &reqs[nreqs]);
            ++nreqs;
            break;
        case CMD_Irecv:
            buff1 = scif_host->getbuf(cmddata.rbuff);
            assert(nreqs < MaxReq);
            assert(ndirty < MaxReq);
            MPI_Irecv(buff1, cmddata.buflen, MPI_BYTE, cmddata.rnode, 0,
                MPI_COMM_WORLD, &reqs[nreqs]);
            ++nreqs;
            dirtybuf[ndirty] = cmddata.rbuff;
            dirtylen[ndirty] = cmddata.buflen;
            ++ndirty;
            break;
        case CMD_Waitall:
            MPI_Waitall(nreqs, reqs, MPI_STATUSES_IGNORE);
            nreqs = 0;
            for (int i = 0; i < ndirty; ++i)
                scif_host->writebuf(dirtybuf[i], dirtylen[i]);
            ndirty = 0;
            scif_host->fence();
            break;
        case CMD_Fence:
            scif_host->fence();
            break;
        case CMD_Sendrecv:
            scif_host->readbuf(cmddata.sbuff, cmddata.buflen);
            scif_host->fence();
            buff1 = scif_host->getbuf(cmddata.sbuff);
            buff2 = scif_host->getbuf(cmddata.rbuff);
            MPI_Sendrecv(buff1, cmddata.buflen, MPI_BYTE, cmddata.snode, 0,
                         buff2, cmddata.buflen, MPI_BYTE, cmddata.rnode, 0,
                         MPI_COMM_WORLD, &status);
            scif_host->writebuf(cmddata.rbuff, cmddata.buflen);
            scif_host->fence();
            break;
        case CMD_Allreduce:
            scif_host->readbuf(cmddata.sbuff, cmddata.buflen);
            scif_host->fence();
            buff1 = scif_host->getbuf(cmddata.sbuff);
            buff2 = scif_host->getbuf(cmddata.rbuff);
            MPI_Allreduce(buff1, buff2, cmddata.buflen/sizeof(float),
                MPI_FLOAT, MPI_SUM, MPI_COMM_WORLD);
            scif_host->writebuf(cmddata.rbuff, cmddata.buflen);
            scif_host->fence();
            break;
        case CMD_Nop:
            break;
        default:
            assert(0);
        }
    }
}
#if 0
__declspec(target(mic)) const int mybz = 128;
__declspec(target(mic))
static char *
norasval(const mic_wqf_eo *qe, int iz, int iy, int ix, char *buf)
{
    int isite = (0 + NTHT*(iz + NZ*(iy + NY*ix)));

    snprintf(buf, mybz, "%12g", ((float *)&qe->s[isite].y[0])[0]);
    return buf;
}
extern "C" void
__declspec(target(mic))
prcorners(char *tag, const mic_wqf_eo *qe)
{
    char buf[mybz];
    printf("%s: X planes\n", tag);
    printf("(   2,   2,   1) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+2-1,NDEPTHX+1-1,buf));
    printf("(NZ-1,   2,   1) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+2-1,NDEPTHX+1-1,buf));
    printf("(   2,NY-1,   1) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+NY-1-1,NDEPTHX+1-1,buf));
    printf("(NZ-1,NY-1,   1) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+NY-1-1,NDEPTHX+1-1,buf));
    printf("(   2,   2,  NX) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+2-1,NDEPTHX+NX-1,buf));
    printf("(NZ-1,   2,  NX) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+2-1,NDEPTHX+NX-1,buf));
    printf("(   2,NY-1,  NX) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+NY-1-1,NDEPTHX+NX-1,buf));
    printf("(NZ-1,NY-1,  NX) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+NY-1-1,NDEPTHX+NX-1,buf));
    printf("%s: Y planes\n", tag);
    printf("(   2,   1,   2) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+1-1,NDEPTHX+2-1,buf));
    printf("(NZ-1,   1,   2) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+1-1,NDEPTHX+2-1,buf));
    printf("(   2,  NY,   2) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+NY-1,NDEPTHX+2-1,buf));
    printf("(NZ-1,  NY,   2) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+NY-1,NDEPTHX+2-1,buf));
    printf("(   2,   1,NX-1) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+1-1,NDEPTHX+NX-1-1,buf));
    printf("(NZ-1,   1,NX-1) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+1-1,NDEPTHX+NX-1-1,buf));
    printf("(   2,  NY,NX-1) %s\n",
      norasval(qe,NDEPTHZ+2-1,NDEPTHY+NY-1,NDEPTHX+NX-1-1,buf));
    printf("(NZ-1,  NY,NX-1) %s\n",
      norasval(qe,NDEPTHZ+NZ-1-1,NDEPTHY+NY-1,NDEPTHX+NX-1-1,buf));
    printf("%s: Z planes\n", tag);
    printf("(   1,   2,   2) %s\n",
      norasval(qe,NDEPTHZ+1-1,NDEPTHY+2-1,NDEPTHX+2-1,buf));
    printf("(  NZ,   2,   2) %s\n",
      norasval(qe,NDEPTHZ+NZ-1,NDEPTHY+2-1,NDEPTHX+2-1,buf));
    printf("(   1,NY-1,   2) %s\n",
      norasval(qe,NDEPTHZ+1-1,NDEPTHY+NY-1-1,NDEPTHX+2-1,buf));
    printf("(  NZ,NY-1,   2) %s\n",
      norasval(qe,NDEPTHZ+NZ-1,NDEPTHY+NY-1-1,NDEPTHX+2-1,buf));
    printf("(   1,   2,NX-1) %s\n",
      norasval(qe,NDEPTHZ+1-1,NDEPTHY+2-1,NDEPTHX+NX-1-1,buf));
    printf("(  NZ,   2,NX-1) %s\n",
      norasval(qe,NDEPTHZ+NZ-1,NDEPTHY+2-1,NDEPTHX+NX-1-1,buf));
    printf("(   1,NY-1,NX-1) %s\n",
      norasval(qe,NDEPTHZ+1-1,NDEPTHY+NY-1-1,NDEPTHX+NX-1-1,buf));
    printf("(  NZ,NY-1,NX-1) %s\n",
      norasval(qe,NDEPTHZ+NZ-1,NDEPTHY+NY-1-1,NDEPTHX+NX-1-1,buf));
}

#endif

extern "C" void
mic_solver(
    const void **fsolver, 
    const float *stol,
    int *iter,
    const mic_wqf_eo *swe,
    mic_wqf_eo *sve)
{
    //prcorners("host solver", swe);
    int offload_signal = 0xF;
    cmic_solver *s = (cmic_solver *)*fsolver;
    const float stol1 = *stol;
    int iter1 = *iter;
    #pragma offload target(mic:mic_targetid) \
        nocopy(scif_mic) \
        in(s[0:0] : alloc_if(0) free_if(0) targetptr) \
        in( swe[0:1] : alloc_if(1) free_if(1)) \
        out(sve[0:1] : alloc_if(1) free_if(1)) \
        inout(iter1) \
        signal(offload_signal)
    {
        float skappa = s->kappa;
        //prcorners("mic solver", swe);
        //fflush(0);
        set_ras_fields(s->u,s->u_ras,s->fclinv,s->fclinv_ras);
        assign_inv_mult_eoprec_wd_bicgstab_mic(&skappa,&stol1,&iter1,sve,swe);
#ifndef _singlePU
        Command cmd;
        cmd.cmd = Command::CMD_Stop;
        cmd.cmdlen = sizeof(Command);
        scif_mic->enqueue(&cmd, sizeof(cmd), true);
        scif_mic->qempty(true);
#endif
    }
    #ifndef _singlePU
    process_cmds();
    #endif
    #pragma offload_wait target(mic:mic_targetid) wait(offload_signal)
    *iter = iter1;
}

extern "C" void
setnodeid(const cmic_solver *s)
{
    #if defined(__MIC__)
    mic_nodeid = s->ni.mic_nodeid;
    mic_nodeid_x_up = s->ni.mic_nodeid_x_up;
    mic_nodeid_y_up = s->ni.mic_nodeid_y_up;
    mic_nodeid_z_up = s->ni.mic_nodeid_z_up;
    mic_nodeid_x_dn = s->ni.mic_nodeid_x_dn;
    mic_nodeid_y_dn = s->ni.mic_nodeid_y_dn;
    mic_nodeid_z_dn = s->ni.mic_nodeid_z_dn;
    mic_nodeid_x = s->ni.mic_nodeid_x;
    mic_nodeid_y = s->ni.mic_nodeid_y;
    mic_nodeid_z = s->ni.mic_nodeid_z;
//    printf("%d: z %d y %d x %d\n", mic_nodeid, mic_nodeid_z, mic_nodeid_y, mic_nodeid_x);
    mic_ipeo = s->ni.mic_ipeo;
    #endif
}
