#ifndef _MIC_MPI_H
#define _MIC_MPI_H

#ifndef _singlePU

// For allreduce
static float *arb_send;
static off_t barb_send;
static float *arb_recv;
static off_t barb_recv;

#include "commands.h"
#include <string.h>

extern "C" static inline void
mic_combuf_wait()
{
    // wait for empty command to ensure all previous commands done
    Command cmd;
    cmd.cmd = CMD_Nop;
    cmd.cmdlen = sizeof(Command);
    scif_mic->enqueue(&cmd, sizeof(cmd), true);
    // wait for host to finish processing
    scif_mic->qempty(true);
}

extern "C" static inline void
mic_combuf_send_start(off_t buf, int count, int dest)
{
    Command cmd[MaxCmd];
    Cmddata cmddata;
    // Tell the host to DMA the send buffer, call isend,
    // and record the request
    // Note: we could start DMA of send buffer first but card DMA
    // setup is so slow it is probably better to just do it on the host
    cmddata.sbuff = buf;
    cmddata.buflen = count;
    cmddata.snode = dest;
    cmd[0].cmd = CMD_Isend;
    cmd[0].cmdlen = sizeof(cmd);
    memcpy(cmd[0].data, &cmddata, sizeof(cmddata));
    scif_mic->enqueue(cmd, sizeof(cmd), true);
}

extern "C" static inline void
mic_combuf_recv_start(off_t buf, int count, int src)
{
    Command cmd[MaxCmd];
    Cmddata cmddata;
    // Tell the host to call irecv, mark the receive buffer dirty,
    // and record the request
    cmddata.rbuff = buf;
    cmddata.buflen = count;
    cmddata.rnode = src;
    cmd[0].cmd = CMD_Irecv;
    cmd[0].cmdlen = sizeof(cmd);
    memcpy(cmd[0].data, &cmddata, sizeof(cmddata));
    scif_mic->enqueue(cmd, sizeof(cmd), true);
}

extern "C" static inline void
mic_combuf_waitall()
{
    Command cmd[MaxCmd];
    Cmddata cmddata;
    // Tell the host to wait for all outstanding requests,
    // dma the dirty recv buffers to the card, and fence the DMAs
    cmd[0].cmd = CMD_Waitall;
    cmd[0].cmdlen = sizeof(Command);
    scif_mic->enqueue(cmd, sizeof(Command), true);
    mic_combuf_wait();
}

extern "C" static inline void
mic_comm_allreduce(float *f1, float *f2, int count)
{
    Command cmd[MaxCmd];
    Cmddata cmddata;
    // Tell the host to do an allreduce:
    //    MPI_Allreduce(f1,f2,count,MPI_FLOAT,MPI_SUM,m_mic_comm);
    //  Return result is in f2
    assert(count <= 2);
    memcpy(arb_send, f1, sizeof(float)*count);
    cmddata.sbuff = barb_send;
    cmddata.rbuff = barb_recv;
    cmddata.buflen = count*sizeof(float);
    cmd[0].cmd = CMD_Allreduce;
    cmd[0].cmdlen = sizeof(cmd);
    memcpy(cmd[0].data, &cmddata, sizeof(cmddata));
    scif_mic->enqueue(cmd, sizeof(cmd), true);
    mic_combuf_wait();  // make sure allreduce done and data returned
    memcpy(f2, arb_recv, sizeof(float)*count);
}

extern "C" static inline void
mic_combuf_sendrecv(int count, off_t sbuf, int snode, off_t rbuf, int rnode)
{
    Command cmd[MaxCmd];
    Cmddata cmddata;
    // Tell the nost to do a sendrecv.
    cmddata.sbuff = sbuf;
    cmddata.snode = snode;
    cmddata.rbuff = rbuf;
    cmddata.rnode = rnode;
    cmddata.buflen = count;
    cmd[0].cmd = CMD_Sendrecv;
    cmd[0].cmdlen = sizeof(cmd);
    memcpy(cmd[0].data, &cmddata, sizeof(cmddata));
    scif_mic->enqueue(cmd, sizeof(cmd), true);
    mic_combuf_wait();  // make sure allreduce done and data returned
}

extern "C" static inline void
mic_combuf_fence()
{
    Command cmd[MaxCmd];
    Cmddata cmddata;
    // Tell the host to fence its DMAs (wait for them to complete)
    cmd[0].cmd = CMD_Fence;
    cmd[0].cmdlen = sizeof(Command);
    scif_mic->enqueue(cmd, sizeof(Command), true);
}

#endif
#endif
