#ifndef _MIC_RAS_SOLVER_CLASS_H
#define _MIC_RAS_SOLVER_CLASS_H

#include "mic_ras_init.h"
#include "copy_boundary_ras_wqf_eo.h"
#include "omp_sched.h"
#include "assign_mult_hop_eo_mic.h"
#include "assign_mult_hop_eo_ras_mic.h"
#include "assign_mult_eoprec_wd_eo_mic.h"
#include "assign_inv_mult_eoprec_wd_ras_mic.h"
#include "mic_ras_solver_class.h"

class  mic_ras_solver {

public:

#if _NRAS > 0 
  static const int is_prec = 1;
#else
  static const int is_prec = 0;
#endif

  static void 
#ifdef _CLOVER
       precon( const float      *skappa,
                     mic_wqf_eo *restrict  ve,
                     mic_wqf_eo *restrict  pe,
               const mic_gvf_soa    *u,
               const mic_clvf   *fclinv);
#else
       precon( const float      *skappa,
                     mic_wqf_eo *restrict  ve,
                     mic_wqf_eo *restrict  pe,
               const mic_gvf_soa   *u);
#endif

  static void 
#ifdef _CLOVER
       matvec( const float      *skappa,
                     mic_wqf_eo *restrict  qe,
                     mic_wqf_eo *restrict  pe,
               const mic_gvf_soa    *u,
               const mic_clvf   *fclinv);
#else
       matvec( const float      *skappa,
                     mic_wqf_eo *restrict  qe,
                     mic_wqf_eo *restrict  pe,
               const mic_gvf_soa    *u);
#endif
};

void
#ifdef _CLOVER
mic_ras_solver::precon( const float      *skappa,
                              mic_wqf_eo *restrict  ve,
                              mic_wqf_eo *restrict  pe,
                        const mic_gvf_soa    *u,
                        const mic_clvf   *fclinv)
{
    assign_inv_mult_eoprec_wd_ras_mic(skappa,ve,pe,u,fclinv);
}
#else
mic_ras_solver::precon( const float      *skappa,
                              mic_wqf_eo *restrict  ve,
                              mic_wqf_eo *restrict  pe,
                        const mic_gvf_soa    *u)
{
    assign_inv_mult_eoprec_wd_ras_mic(skappa,ve,pe,u);
}
#endif

void
#ifdef _CLOVER
mic_ras_solver::matvec( const float      *skappa,
                              mic_wqf_eo *restrict  qe,
                              mic_wqf_eo *restrict  pe,
                        const mic_gvf_soa    *u,
                        const mic_clvf   *fclinv)
{
    assign_mult_eoprec_wd_eo_mic(skappa,qe,pe,u,fclinv);
}
#else
mic_ras_solver::matvec( const float      *skappa,
                              mic_wqf_eo *restrict  qe,
                              mic_wqf_eo *restrict  pe,
                        const mic_gvf_soa    *u)
{
    assign_mult_eoprec_wd_eo_mic(skappa,qe,pe,u);
}
#endif

extern "C" void
assign_inv_mult_eoprec_wd_bicgstab_mic(const float  *skappa,
                                       const float  *stol,
                                       int          *iter,
                                             mic_wqf_eo   *restrict yde,
                                       const mic_wqf_eo   *restrict  ye)
{
  mic_ras_solver solver;
  //prcorners("bicgstab", ye);
#ifdef _CLOVER
  assign_inv_matvec(solver,skappa,stol,iter,yde,ye,mic_u,mic_fclinv);
#else
  assign_inv_matvec(solver,skappa,stol,iter,yde,ye,mic_u);
#endif
}

#endif
