#include "config.h"
#include "mic_config.h"

module mic_single_field_ras_class
!
! Single precision solver API used in the host program.
!
  use ISO_C_BINDING
  use lattice_class
  implicit none
  private

  public :: half_spinor
  public :: mic_su3_spinor
  public :: mic_wqf_eo
  public :: mic_wqf
  public :: mic_ras_wqf_eo
  public :: mic_ras_wqf
  public :: su3mat
  public :: mic_gsf_eo, mic_gsf_eo_soa
  public :: mic_gvf_eo, mic_gvf_eo_soa
  public :: mic_gvf, mic_gvf_soa
  public :: mic_ras_gsf_eo, mic_ras_gsf_eo_soa
  public :: mic_ras_gvf_eo, mic_ras_gvf_eo_soa
  public :: mic_ras_gvf, mic_ras_gvf_soa
  public :: half_clover_term
  public :: mic_clover
  public :: mic_clvf_eo
  public :: mic_clvf
  public :: mic_ras_clvf_eo
  public :: mic_ras_clvf
!  public :: conv_q_d2s_eo
!  public :: conv_q_s2d_eo
!  public :: conv_u_d2s_eo
!  public :: conv_fclv_d2s_eo
!  public :: assign_mult_hop_eo_mic
!  public :: assign_mult_eoprec_wd_eo_mic
  public :: mic_ras_solver
  public :: assign_inv_mult_eoprec_wd
  public :: new
  public :: delete

  public :: copy_boundary_ras_wqf_eo


  integer(C_INT), parameter :: COMMAND_NEW    = 0
  integer(C_INT), parameter :: COMMAND_SOLVE  = 1
  integer(C_INT), parameter :: COMMAND_DELETE = 2


  !========================================
  ! RAS overlap region depth
  !========================================
#if _NDIMZ != 1
  integer, parameter :: NDEPTHZ = (_NDEPTH)
#else
  integer, parameter :: NDEPTHZ = 0
#endif
#if _NDIMY != 1
  integer, parameter :: NDEPTHY = (_NDEPTH)
#else
  integer, parameter :: NDEPTHY = 0
#endif
#if _NDIMX != 1
  integer, parameter :: NDEPTHX = (_NDEPTH)
#else
  integer, parameter :: NDEPTHX = 0
#endif

  integer, parameter :: SP=C_FLOAT
  integer, parameter :: DP=KIND(1.0d0)

  integer, parameter :: VLEN=4
  character(len=*), parameter :: archname="ARCH=MIC-OFFLOAD:"

  integer, parameter :: NTHT=NTH/VLEN

  real(DP), parameter :: MAX_TOL = 0.01_DP
  real(DP), parameter :: MIN_TOL = 0.00001_DP

  !========================================
  ! node information for C-lang solver
  !========================================
  type, BIND(C) :: nodeinfo
    integer(C_INT) :: mic_nodeid
    integer(C_INT) :: mic_nodeid_x_up
    integer(C_INT) :: mic_nodeid_y_up
    integer(C_INT) :: mic_nodeid_z_up
    integer(C_INT) :: mic_nodeid_x_dn
    integer(C_INT) :: mic_nodeid_y_dn
    integer(C_INT) :: mic_nodeid_z_dn
    integer(C_INT) :: mic_nodeid_x
    integer(C_INT) :: mic_nodeid_y
    integer(C_INT) :: mic_nodeid_z
    integer(C_INT) :: mic_ipdeo;
    integer(C_INT) :: mic_ipeo
  end type

  type(nodeinfo), BIND(C), save :: m_host_nodeinfo
  integer, save :: m_is_mic_nodeinfo_initialized = .FALSE.

  !=================================
  ! module initialization status
  !=================================
  integer, save :: m_is_mic_initialized = .FALSE.

  !==========================
  ! MPI buffer
  !==========================
  type buffer_obj
    complex(4), pointer :: buff_up(:,:) 
    complex(4), pointer :: buff_dn(:,:) 
    type(comlib_data_c8) :: id_sendup
    type(comlib_data_c8) :: id_senddn
    integer :: size = 0
    integer :: dummy1 = 0
    integer :: dummy2 = 0
    integer :: dummy3 = 0
  end type 
  public :: buffer_obJ

  type(buffer_obj), public, save :: wqhf_buff(NDIM-1)
  type(buffer_obj), public, save ::  wqf_buff(NDIM-1)
  type(buffer_obj), public, save ::  gvf_buff(NDIM-1)
  type(buffer_obj), public, save :: clvf_buff(NDIM-1)

  !==================================
  ! Field Definitions
  !==================================
  type, BIND(C) :: half_spinor
    real(SP) :: y1r(VLEN)
    real(SP) :: y2r(VLEN)
    real(SP) :: y1i(VLEN)
    real(SP) :: y2i(VLEN)
  end type

  type, BIND(C) :: mic_su3_spinor
    type(half_spinor) :: y(COL,2)
  end type

  type, BIND(C) :: mic_wqf_eo
    type(mic_su3_spinor) :: s(NTHT,NZ,NY,NX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_ras_wqf_eo
    type(mic_su3_spinor) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_wqf
    type(mic_wqf_eo) :: eo(0:1)
  end type

  type, BIND(C) :: mic_ras_wqf
    type(mic_ras_wqf_eo) :: eo(0:1)
  end type

  type, BIND(C) :: su3mat
    real(SP) :: u11r(VLEN)
    real(SP) :: u12r(VLEN)
    real(SP) :: u11i(VLEN)
    real(SP) :: u12i(VLEN)

    real(SP) :: u21r(VLEN)
    real(SP) :: u22r(VLEN)
    real(SP) :: u21i(VLEN)
    real(SP) :: u22i(VLEN)

    real(SP) :: u31r(VLEN)
    real(SP) :: u32r(VLEN)
    real(SP) :: u31i(VLEN)
    real(SP) :: u32i(VLEN)
  end type

  type, BIND(C) :: mic_gsf_eo
    type(su3mat) :: s(NTHT,NZ,NY,NX)
  end type
  type, BIND(C) :: mic_gsf_eo_soa
    type(su3mat) :: v(NDIM)
  end type

  type, BIND(C) :: mic_ras_gsf_eo
    type(su3mat) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
  end type
  type, BIND(C) :: mic_ras_gsf_eo_soa
    type(su3mat) :: v(NDIM)
  end type

  type, BIND(C) :: mic_gvf_eo
    type(mic_gsf_eo) :: v(NDIM)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type
  type, BIND(C) :: mic_gvf_eo_soa
    type(mic_gsf_eo_soa) :: s(NTHT,NZ,NY,NX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_ras_gvf_eo
    type(mic_ras_gsf_eo) :: v(NDIM)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type
  type, BIND(C) :: mic_ras_gvf_eo_soa
    type(mic_ras_gsf_eo_soa) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_gvf
    type(mic_gvf_eo) :: eo(0:1)
  end type
  type, BIND(C) :: mic_gvf_soa
    type(mic_gvf_eo_soa) :: eo(0:1)
  end type

  type, BIND(C) :: mic_ras_gvf
    type(mic_ras_gvf_eo) :: eo(0:1)
  end type
  type, BIND(C) :: mic_ras_gvf_soa
    type(mic_ras_gvf_eo_soa) :: eo(0:1)
  end type

  type, BIND(C) :: half_clover_term
    real(SP) ::  f1r(VLEN)
    real(SP) :: f16r(VLEN)
    real(SP) ::  f4r(VLEN)
    real(SP) ::  f4i(VLEN)

    real(SP) ::  f2r(VLEN)
    real(SP) :: f17r(VLEN)
    real(SP) ::  f2i(VLEN)
    real(SP) :: f17i(VLEN)

    real(SP) ::  f5r(VLEN)
    real(SP) ::  f9r(VLEN)
    real(SP) ::  f5i(VLEN)
    real(SP) ::  f9i(VLEN)

    real(SP) ::  f3r(VLEN)
    real(SP) :: f18r(VLEN)
    real(SP) ::  f3i(VLEN)
    real(SP) :: f18i(VLEN)

    real(SP) ::  f6r(VLEN)
    real(SP) :: f13r(VLEN)
    real(SP) ::  f6i(VLEN)
    real(SP) :: f13i(VLEN)

    real(SP) ::  f7r(VLEN)
    real(SP) :: f19r(VLEN)
    real(SP) :: f10r(VLEN)
    real(SP) :: f10i(VLEN)

    real(SP) ::  f8r(VLEN)
    real(SP) :: f20r(VLEN)
    real(SP) ::  f8i(VLEN)
    real(SP) :: f20i(VLEN)

    real(SP) :: f11r(VLEN)
    real(SP) :: f14r(VLEN)
    real(SP) :: f11i(VLEN)
    real(SP) :: f14i(VLEN)

    real(SP) :: f12r(VLEN)
    real(SP) :: f21r(VLEN)
    real(SP) :: f15r(VLEN)
    real(SP) :: f15i(VLEN)
  end type

  type, BIND(C) :: mic_clover
    type(half_clover_term) :: c(2)
  end type

  type, BIND(C) :: mic_clvf_eo
    type(mic_clover) :: s(NTHT,NZ,NY,NX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_ras_clvf_eo
    type(mic_clover) :: s(NTHT,1-NDEPTHZ:NZ+NDEPTHZ,1-NDEPTHY:NY+NDEPTHY,1-NDEPTHX:NX+NDEPTHX)
    integer(C_INT) :: ieo
    integer(C_INT) :: idummy(15)
  end type

  type, BIND(C) :: mic_clvf
    type(mic_clvf_eo) :: eo(0:1)
  end type

  type, BIND(C) :: mic_ras_clvf
    type(mic_ras_clvf_eo) :: eo(0:1)
  end type

  !======================================
  ! MIC Kernel solver
  !======================================
  type mic_ras_solver
    private
    type(mic_gvf_soa),    allocatable :: u
!DIR$ ATTRIBUTES ALIGN:64 :: u

    type(mic_ras_gvf_soa),    allocatable :: u_ras
!DIR$ ATTRIBUTES ALIGN:64 :: u_ras

    type(mic_clvf),   allocatable :: fclinv
!DIR$ ATTRIBUTES ALIGN:64 :: fclinv

    type(mic_ras_clvf),   allocatable :: fclinv_ras
!DIR$ ATTRIBUTES ALIGN:64 :: fclinv_ras

    type(mic_wqf_eo), allocatable :: swe
!DIR$ ATTRIBUTES ALIGN:64 :: swe

    type(mic_wqf_eo), allocatable :: sve
!DIR$ ATTRIBUTES ALIGN:64 :: sve
    real(DP), public :: kappa = 0.0_DP
    real(DP), public :: err_norm =0.0_DP
    real(DP), public :: tol = 1.0e-4_DP
    integer,  public :: iter = 0
    integer,  public :: print_flag = 0
    integer,  public :: ieo = 0
  end type

  interface

    subroutine new_offl_ras_solver(u,u_ras,fclinv,fclinv_ras) BIND(C)
      import
      type(mic_gvf_soa), intent(in):: u
      type(mic_ras_gvf_soa), intent(in):: u_ras
      type(mic_clvf), intent(in):: fclinv
      type(mic_ras_clvf), intent(in):: fclinv_ras
    end subroutine

    subroutine delete_offl_ras_solver() BIND(C)
    end subroutine

    subroutine assign_inv_mult_offl_ras_solver(skappa,stol,iter,swe,sve) BIND(C)
      import
      real(SP),         intent(in)    :: skappa
      integer(C_INT),   intent(inout) :: iter
      type(mic_wqf_eo), intent(in)    :: swe
      type(mic_wqf_eo), intent(out)   :: sve
    end subroutine

  end interface

  interface new
    module procedure new_mic
  end interface

  interface delete
    module procedure delete_mic
  end interface

  interface assign_inv_mult_eoprec_wd
    module procedure assign_inv_mult_eoprec_wd_mic
  end interface

contains

subroutine new_mic(this,kappa,ue,uo,ieo,fclinve,fclinvo)
  implicit none
  type(mic_ras_solver), intent(inout) :: this
  real(DP),         intent(in)    :: kappa
  complex(DP),      intent(in)    :: ue(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(DP),      intent(in)    :: uo(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  integer,          intent(in)    :: ieo
  complex(DP),      intent(in)    :: fclinve(CLSPH,0:NTH,NZ,NY,NX,2)
  complex(DP),      intent(in)    :: fclinvo(CLSPH,0:NTH,NZ,NY,NX,2)
  integer :: ioe
  integer :: mu, nthmodvlen

  m_host_nodeinfo%mic_nodeid = nodeid
  m_host_nodeinfo%mic_nodeid_x = ipsite(1)
  m_host_nodeinfo%mic_nodeid_y = ipsite(2)
  m_host_nodeinfo%mic_nodeid_z = ipsite(3)
  m_host_nodeinfo%mic_nodeid_x_up = nodeidup(1)
  m_host_nodeinfo%mic_nodeid_y_up = nodeidup(2)
  m_host_nodeinfo%mic_nodeid_z_up = nodeidup(3)
  m_host_nodeinfo%mic_nodeid_x_dn = nodeiddn(1)
  m_host_nodeinfo%mic_nodeid_y_dn = nodeiddn(2)
  m_host_nodeinfo%mic_nodeid_z_dn = nodeiddn(3)
  m_host_nodeinfo%mic_ipeo   = ipeo

  if (0 == nodeid) then
    write(*,'("%NDEPTHZYX:",3I3)') NDEPTHZ,NDEPTHY,NDEPTHX
  endif

  nthmodvlen = mod(NTH,VLEN)
  if ( 0 == nodeid ) then
    write(*,'("%",A)')archname
    write(*,'("%VLEN=",I3)')VLEN
    write(*,'("%NTH % VLEN=",I3)') nthmodvlen
  endif
  if ( 0 /= nthmodvlen ) then
    if ( 0 == nodeid ) then
      write(*,'("Stop by NT error! NTT must be a multple of (VLEN*2).")')
    endif
#ifndef _singlePU
    call comlib_finalize()
#endif
    stop
  endif

#ifndef _singlePU

#if _NDIMZ != 1
   wqf_buff(3)%size = CLSP         *VLEN*NTHT*NY*NX*NDEPTHZ
   gvf_buff(3)%size = COL*(COL-1)  *VLEN*NTHT*NY*NX*NDEPTHZ*NDIM
  clvf_buff(3)%size = (CLSPH-COL)*2*VLEN*NTHT*NY*NX*NDEPTHZ
#endif
#if _NDIMY != 1
   wqf_buff(2)%size = CLSP         *VLEN*NTHT*(NZ+NDEPTHZ*2)*NX*NDEPTHY
   gvf_buff(2)%size = COL*(COL-1)  *VLEN*NTHT*(NZ+NDEPTHZ*2)*NX*NDEPTHY*NDIM
  clvf_buff(2)%size = (CLSPH-COL)*2*VLEN*NTHT*(NZ+NDEPTHZ*2)*NX*NDEPTHY
#endif
#if _NDIMX != 1
   wqf_buff(1)%size = CLSP         *VLEN*NTHT*(NZ+NDEPTHZ*2)*(NY+NDEPTHY*2)*NDEPTHX
   gvf_buff(1)%size = COL*(COL-1)  *VLEN*NTHT*(NZ+NDEPTHZ*2)*(NY+NDEPTHY*2)*NDEPTHX*NDIM
  clvf_buff(1)%size = (CLSPH-COL)*2*VLEN*NTHT*(NZ+NDEPTHZ*2)*(NY+NDEPTHY*2)*NDEPTHX
#endif

  do mu=1,NDIM-1
    allocate( wqf_buff(mu)%buff_up( wqf_buff(mu)%size,2))
    allocate( wqf_buff(mu)%buff_dn( wqf_buff(mu)%size,2))
    allocate( gvf_buff(mu)%buff_up( gvf_buff(mu)%size,2))
    allocate( gvf_buff(mu)%buff_dn( gvf_buff(mu)%size,2))
    allocate(clvf_buff(mu)%buff_up(clvf_buff(mu)%size,2))
    allocate(clvf_buff(mu)%buff_dn(clvf_buff(mu)%size,2))
  enddo

  do mu=1,NDIM-1
    call comlib_make2(wqf_buff(mu)%id_sendup,nodeidup(mu),  &
                      wqf_buff(mu)%buff_up(1,1),  &
                      wqf_buff(mu)%buff_up(1,2),wqf_buff(mu)%size*8)
    call comlib_make2(wqf_buff(mu)%id_senddn,nodeiddn(mu),  &
                      wqf_buff(mu)%buff_dn(1,1),  &
                      wqf_buff(mu)%buff_dn(1,2),wqf_buff(mu)%size*8)

    call comlib_make2(gvf_buff(mu)%id_sendup,nodeidup(mu),  &
                      gvf_buff(mu)%buff_up(1,1),  &
                      gvf_buff(mu)%buff_up(1,2),gvf_buff(mu)%size*8)
    call comlib_make2(gvf_buff(mu)%id_senddn,nodeiddn(mu),  &
                      gvf_buff(mu)%buff_dn(1,1),  &
                      gvf_buff(mu)%buff_dn(1,2),gvf_buff(mu)%size*8)

    call comlib_make2(clvf_buff(mu)%id_sendup,nodeidup(mu),  &
                      clvf_buff(mu)%buff_up(1,1),  &
                      clvf_buff(mu)%buff_up(1,2),clvf_buff(mu)%size*8)
    call comlib_make2(clvf_buff(mu)%id_senddn,nodeiddn(mu),  &
                      clvf_buff(mu)%buff_dn(1,1),  &
                      clvf_buff(mu)%buff_dn(1,2),clvf_buff(mu)%size*8)
  enddo

#endif

  if (allocated(this%u))          deallocate(this%u)
  if (allocated(this%u_ras))      deallocate(this%u_ras)
#ifdef _CLOVER
  if (allocated(this%fclinv))     deallocate(this%fclinv)
  if (allocated(this%fclinv_ras)) deallocate(this%fclinv_ras)
#endif
  if (allocated(this%swe))    deallocate(this%swe)
  if (allocated(this%sve))    deallocate(this%sve)

  allocate(this%u)
  allocate(this%u_ras)
#ifdef _CLOVER
  allocate(this%fclinv)
  allocate(this%fclinv_ras)
#endif
  allocate(this%swe)
  allocate(this%sve)

  ioe = 1 - ieo
  this%ieo   = ieo
  this%iter  = 0
  this%err_norm = 0.0_DP
  this%tol      = 1.0_DP
  this%kappa    = kappa

  !===========================================
  !  send data to mic
  !===========================================

  call conv_u_d2s_eo(this%u%eo(ieo),ue,ieo)
  call conv_u_d2s_eo(this%u%eo(ioe),uo,ioe)
  call assign_u_ras_eo(this%u_ras%eo(ieo),this%u%eo(ieo))
  call assign_u_ras_eo(this%u_ras%eo(ioe),this%u%eo(ioe))

#ifdef _CLOVER
  call conv_fclv_d2s_eo(this%fclinv%eo(ieo),fclinve,ieo)
  call conv_fclv_d2s_eo(this%fclinv%eo(ioe),fclinvo,ioe)
  call assign_fclv_ras_eo(this%fclinv_ras%eo(ieo),this%fclinv%eo(ieo))
  call assign_fclv_ras_eo(this%fclinv_ras%eo(ioe),this%fclinv%eo(ioe))
#endif

  call new_offl_ras_solver( this%u, this%u_ras, this%fclinv, this%fclinv_ras);
  m_is_mic_initialized = .TRUE.

  return
end subroutine

subroutine delete_mic(this)
  implicit none
  type(mic_ras_solver), intent(inout) :: this
  integer :: mu

  if (m_is_mic_initialized) then

    call delete_offl_ras_solver()
#ifndef _singlePU
    do mu=1,NDIM-1
      if (associated( wqf_buff(mu)%buff_up)) deallocate( wqf_buff(mu)%buff_up)
      if (associated( wqf_buff(mu)%buff_dn)) deallocate( wqf_buff(mu)%buff_dn)
      if (associated( gvf_buff(mu)%buff_up)) deallocate( gvf_buff(mu)%buff_up)
      if (associated( gvf_buff(mu)%buff_dn)) deallocate( gvf_buff(mu)%buff_dn)
      if (associated(clvf_buff(mu)%buff_up)) deallocate(clvf_buff(mu)%buff_up)
      if (associated(clvf_buff(mu)%buff_dn)) deallocate(clvf_buff(mu)%buff_dn)
    enddo
#endif

    if (allocated(this%u))      deallocate(this%u)
    if (allocated(this%fclinv)) deallocate(this%fclinv)
    if (allocated(this%swe))    deallocate(this%swe)
    if (allocated(this%sve))    deallocate(this%sve)
#ifndef _singlePU
    if (allocated(this%u_ras))      deallocate(this%u_ras)
    if (allocated(this%fclinv_ras)) deallocate(this%fclinv_ras)
#endif

    m_is_mic_initialized = .FALSE.

  endif

  return
end subroutine

subroutine assign_inv_mult_eoprec_wd_mic(this,dye,ye)
  use omp_lib
  implicit none
  type(mic_ras_solver), intent(inout) :: this
  complex(DP),      intent(inout) :: dye(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(DP),      intent(in)    ::  ye(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  real(DP) :: t0,t1
  real(DP) :: rnorm
  real(SP) :: stol,skappa
  integer(C_INT)  :: ieo,iter
  t0 = omp_get_wtime()

  skappa = this%kappa
  iter = this%iter
  stol = MAX(MIN(this%tol/this%err_norm/10.0_DP,MAX_TOL),MIN_TOL)
  ieo  = this%ieo

  call conv_q_d2s_eo(this%swe,ye,ieo,rnorm)

  call assign_inv_mult_offl_ras_solver(skappa,stol,iter,this%swe,this%sve)
  this%iter = iter

  call conv_q_s2d_eo(dye,this%sve,rnorm)

  if (this%iter < 0) then
    call delete(this)
    if (0 == nodeid) then
      write(*,'("%assign_inv_mult_eoprec_wd_bicgstab_mic : maximum iteration reached.")')
    endif
  endif

  t1 = omp_get_wtime()
  if (0 == nodeid) then
    write(*,'("#ETIME_=",ES24.15)')(t1-t0)
  endif

  return
end subroutine

subroutine conv_q_d2s_eo(sqe,dqe,ieo,rnorm)
  implicit none
  type(mic_wqf_eo), intent(inout) :: sqe
  complex(DP),      intent(in) :: dqe(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  integer,          intent(in) :: ieo
  real(DP), optional, intent(out) :: rnorm
  integer :: ix,iy,iz,ieoxyz,ith,itht,ivec
  integer :: ic,is
  real(DP) :: rtmp

  sqe%ieo  = ieo

  if (present(rnorm)) then
    rnorm = 0.0d0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ic,is) COLLAPSE(3) REDUCTION(+:rnorm)
    do ix=1,NX
    do iy=1,NY
    do iz=1,NZ
      ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
      do ith=1-ieoxyz,NTH-ieoxyz
        do is=1,SPIN
        do ic=1,COL
          rnorm = rnorm +  real(dqe(ic,is,ith,iz,iy,ix),kind=KIND(1.0d0))**2 &
 &                      + aimag(dqe(ic,is,ith,iz,iy,ix))**2
        enddo
        enddo
      enddo
    enddo
    enddo
    enddo
#ifndef _singlePU
    call comlib_sumcast(rnorm)
#endif
    rnorm = SQRT(rnorm)
    rtmp = 1.0d0/rnorm
  else
    rtmp = 1.0d0
  endif
!  print *,'rnorm,rtmp',rnorm,rtmp

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ic,itht,ivec) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1
      do ic=1,COL
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y1r(ivec) =  real(dqe(ic,1,ith,iz,iy,ix),kind=KIND(1.0d0))*rtmp
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y1i(ivec) = aimag(dqe(ic,1,ith,iz,iy,ix))*rtmp
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y2r(ivec) =  real(dqe(ic,2,ith,iz,iy,ix),kind=KIND(1.0d0))*rtmp
        sqe%s(itht,iz,iy,ix)%y(ic,1)%y2i(ivec) = aimag(dqe(ic,2,ith,iz,iy,ix))*rtmp
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y1r(ivec) =  real(dqe(ic,3,ith,iz,iy,ix),kind=KIND(1.0d0))*rtmp
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y1i(ivec) = aimag(dqe(ic,3,ith,iz,iy,ix))*rtmp
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y2r(ivec) =  real(dqe(ic,4,ith,iz,iy,ix),kind=KIND(1.0d0))*rtmp
        sqe%s(itht,iz,iy,ix)%y(ic,2)%y2i(ivec) = aimag(dqe(ic,4,ith,iz,iy,ix))*rtmp
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine conv_q_s2d_eo(dqe,sqe,rnorm)
  implicit none
  complex(DP),      intent(inout) :: dqe(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  type(mic_wqf_eo), intent(in) :: sqe
  real(DP),  optional, intent(in) :: rnorm
  integer :: ieo
  integer :: ix,iy,iz,ieoxyz,ith,itht,ivec
  integer :: ic,is
  complex(DP) :: ctmp
  real(DP) :: rtmp

  ieo = sqe%ieo

  if (present(rnorm)) then
    rtmp = rnorm
  else
    rtmp = 1.0d0
  endif

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,ic,itht,ivec) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1
      do ic=1,COL
        dqe(ic,1,ith,iz,iy,ix) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,1)%y1r(ivec),  &
  &                                    sqe%s(itht,iz,iy,ix)%y(ic,1)%y1i(ivec),kind=DP)*rtmp
        dqe(ic,2,ith,iz,iy,ix) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,1)%y2r(ivec),  &
  &                                    sqe%s(itht,iz,iy,ix)%y(ic,1)%y2i(ivec),kind=DP)*rtmp
        dqe(ic,3,ith,iz,iy,ix) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,2)%y1r(ivec),  &
  &                                    sqe%s(itht,iz,iy,ix)%y(ic,2)%y1i(ivec),kind=DP)*rtmp
        dqe(ic,4,ith,iz,iy,ix) = CMPLX(sqe%s(itht,iz,iy,ix)%y(ic,2)%y2r(ivec),  &
  &                                    sqe%s(itht,iz,iy,ix)%y(ic,2)%y2i(ivec),kind=DP)*rtmp
      enddo
    enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine conv_u_d2s_eo(sue,due,ieo)
  implicit none
  type(mic_gvf_eo_soa), intent(inout) :: sue
  complex(DP),      intent(in)    :: due(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  integer,          intent(in)    :: ieo
  integer :: ix,iy,iz,ieoxyz,ith,itht,ivec
  integer :: ic,jc,mu

  sue%ieo  = ieo

!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,ieoxyz,ith,itht,ivec) COLLAPSE(4)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
  do mu=1,NDIM
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1

      sue%s(itht,iz,iy,ix)%v(mu)%u11r(ivec) =  real(due(1,1,ith,iz,iy,ix,mu),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u11i(ivec) = aimag(due(1,1,ith,iz,iy,ix,mu))
      sue%s(itht,iz,iy,ix)%v(mu)%u21r(ivec) =  real(due(2,1,ith,iz,iy,ix,mu),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u21i(ivec) = aimag(due(2,1,ith,iz,iy,ix,mu))
      sue%s(itht,iz,iy,ix)%v(mu)%u31r(ivec) =  real(due(3,1,ith,iz,iy,ix,mu),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u31i(ivec) = aimag(due(3,1,ith,iz,iy,ix,mu))
      sue%s(itht,iz,iy,ix)%v(mu)%u12r(ivec) =  real(due(1,2,ith,iz,iy,ix,mu),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u12i(ivec) = aimag(due(1,2,ith,iz,iy,ix,mu))
      sue%s(itht,iz,iy,ix)%v(mu)%u22r(ivec) =  real(due(2,2,ith,iz,iy,ix,mu),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u22i(ivec) = aimag(due(2,2,ith,iz,iy,ix,mu))
      sue%s(itht,iz,iy,ix)%v(mu)%u32r(ivec) =  real(due(3,2,ith,iz,iy,ix,mu),kind=DP)
      sue%s(itht,iz,iy,ix)%v(mu)%u32i(ivec) = aimag(due(3,2,ith,iz,iy,ix,mu))

    enddo
  enddo
  enddo
  enddo
  enddo

  return
end subroutine

subroutine assign_u_ras_eo(ras_ue,ue)
  implicit none
  type(mic_ras_gvf_eo_soa), intent(inout) :: ras_ue
  type(mic_gvf_eo_soa),     intent(in)    :: ue
  integer :: ix,iy,iz,itht
  integer :: ic,jc,mu
  integer :: ieo

  ieo  = ue%ieo
  ras_ue%ieo  = ieo

!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itht=1,NTHT
      ras_ue%s(itht,iz,iy,ix)%v(mu) = ue%s(itht,iz,iy,ix)%v(mu)
    enddo
  enddo
  enddo
  enddo
  enddo

  call copy_boundary_ras_gvf_eo(ras_ue)

  return
end subroutine


subroutine conv_fclv_d2s_eo(sfclve,dfclve,ieo)
  implicit none
  type(mic_clvf_eo), intent(inout) :: sfclve
  complex(DP),       intent(in)    :: dfclve(CLSPH,0:NTH,NZ,NY,NX,2)
  integer,           intent(in)    :: ieo
  integer :: ix,iy,iz,ieoxyz,ith,itht,ivec
  integer :: is

  sfclve%ieo  = ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,ieoxyz,ith,itht,ivec,is) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    ieoxyz=mod(ipeo+ix+iy+iz+ieo,2)
    do ith=1-ieoxyz,NTH-ieoxyz
      ivec = mod((ith+ieoxyz-1),VLEN)+1
      itht =     (ith+ieoxyz-1)/VLEN +1

      do is=1,2
        sfclve%s(itht,iz,iy,ix)%c(is)%f1r(ivec)  =  real(dfclve( 1,ith,iz,iy,ix,is),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f2r(ivec)  =  real(dfclve( 2,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f2i(ivec)  = aimag(dfclve( 2,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f3r(ivec)  =  real(dfclve( 3,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f3i(ivec)  = aimag(dfclve( 3,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f4r(ivec)  =  real(dfclve( 4,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f4i(ivec)  = aimag(dfclve( 4,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f5r(ivec)  =  real(dfclve( 5,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f5i(ivec)  = aimag(dfclve( 5,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f6r(ivec)  =  real(dfclve( 6,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f6i(ivec)  = aimag(dfclve( 6,ith,iz,iy,ix,is))

        sfclve%s(itht,iz,iy,ix)%c(is)%f7r(ivec)  =  real(dfclve( 7,ith,iz,iy,ix,is),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f8r(ivec)  =  real(dfclve( 8,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f8i(ivec)  = aimag(dfclve( 8,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f9r(ivec)  =  real(dfclve( 9,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f9i(ivec)  = aimag(dfclve( 9,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f10r(ivec) =  real(dfclve(10,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f10i(ivec) = aimag(dfclve(10,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f11r(ivec) =  real(dfclve(11,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f11i(ivec) = aimag(dfclve(11,ith,iz,iy,ix,is))

        sfclve%s(itht,iz,iy,ix)%c(is)%f12r(ivec) =  real(dfclve(12,ith,iz,iy,ix,is),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f13r(ivec) =  real(dfclve(13,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f13i(ivec) = aimag(dfclve(13,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f14r(ivec) =  real(dfclve(14,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f14i(ivec) = aimag(dfclve(14,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f15r(ivec) =  real(dfclve(15,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f15i(ivec) = aimag(dfclve(15,ith,iz,iy,ix,is))

        sfclve%s(itht,iz,iy,ix)%c(is)%f16r(ivec) =  real(dfclve(16,ith,iz,iy,ix,is),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f17r(ivec) =  real(dfclve(17,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f17i(ivec) = aimag(dfclve(17,ith,iz,iy,ix,is))
        sfclve%s(itht,iz,iy,ix)%c(is)%f18r(ivec) =  real(dfclve(18,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f18i(ivec) = aimag(dfclve(18,ith,iz,iy,ix,is))

        sfclve%s(itht,iz,iy,ix)%c(is)%f19r(ivec) =  real(dfclve(19,ith,iz,iy,ix,is),kind=DP)
!=========
        sfclve%s(itht,iz,iy,ix)%c(is)%f20r(ivec) =  real(dfclve(20,ith,iz,iy,ix,is),kind=DP)
        sfclve%s(itht,iz,iy,ix)%c(is)%f20i(ivec) = aimag(dfclve(20,ith,iz,iy,ix,is))

        sfclve%s(itht,iz,iy,ix)%c(is)%f21r(ivec) =  real(dfclve(21,ith,iz,iy,ix,is),kind=DP)
!=========

      enddo

    enddo
  enddo
  enddo
  enddo

  return
end subroutine


subroutine assign_fclv_ras_eo(ras_fclve,fclve)
  implicit none
  type(mic_ras_clvf_eo), intent(inout) :: ras_fclve
  type(mic_clvf_eo),     intent(in)    :: fclve
  integer :: ix,iy,iz,itht,ieo
  integer :: is

  ieo  = fclve%ieo
  ras_fclve%ieo  = ieo

!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht) COLLAPSE(4)
  do ix=1,NX
  do iy=1,NY
  do iz=1,NZ
    do itht=1,NTHT
      ras_fclve%s(itht,iz,iy,ix) = fclve%s(itht,iz,iy,ix)
    enddo
  enddo
  enddo
  enddo

  call copy_boundary_ras_clvf_eo(ras_fclve)

  return
end subroutine

subroutine copy_boundary_ras_wqf_eo(qe) BIND(C)
  type(mic_ras_wqf_eo), intent(inout) :: qe
  integer :: ix,iy,iz,itht,ivec,ic,is,ibuff
  integer, parameter :: NDAT = VLEN*COL*SPIN
  complex(4) :: ctmp(NDAT)
  type(mic_su3_spinor) :: su3tmp

#if _NDIMZ != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      wqf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,NZ-NDEPTHZ+1+iz,iy,ix),mold=ctmp)
      wqf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,           1+iz,iy,ix),mold=ctmp)

    enddo
  enddo
  enddo
  enddo
!  ibuff = NDAT*NDEPTHZ*NTHT*NX*NY
!  if (0 == nodeid) write(*,'("F:Z-DIR SENDRECV SIZE:",I10,I10)')ibuff,wqf_buff(3)%size
  call comlib_sendrecv(wqf_buff(3)%id_sendup)
  call comlib_sendrecv(wqf_buff(3)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      qe%s(itht,1-NDEPTHZ+iz,iy,ix) = TRANSFER(wqf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      qe%s(itht,     NZ+1+iz,iy,ix) = TRANSFER(wqf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)

    enddo
  enddo
  enddo
  enddo
#endif
  
#if _NDIMY != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      wqf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,NY-NDEPTHY+1+iy,ix),mold=ctmp)
      wqf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,           1+iy,ix),mold=ctmp)

    enddo
  enddo
  enddo
  enddo
!  ibuff = NDAT*NDEPTHY*NTHT*NZ*(2*NDEPTHZ + NZ)
!  if (0 == nodeid) write(*,'("F:Y-DIR SENDRECV SIZE:",I10,I10)')ibuff,wqf_buff(2)%size
  call comlib_sendrecv(wqf_buff(2)%id_sendup)
  call comlib_sendrecv(wqf_buff(2)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      qe%s(itht,iz,1-NDEPTHY+iy,ix) = TRANSFER(wqf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      qe%s(itht,iz,     NY+1+iy,ix) = TRANSFER(wqf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)

    enddo
  enddo
  enddo
  enddo
#endif

#if _NDIMX != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      wqf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,iy,NX-NDEPTHX+1+ix),mold=ctmp)
      wqf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(qe%s(itht,iz,iy,           1+ix),mold=ctmp)

    enddo
  enddo
  enddo
  enddo
!  ibuff = NDAT*NDEPTHX*NTHT*(2*NDEPTHY + NY)*(2*NDEPTHZ + NZ)
!  if (0 == nodeid) write(*,'("F:X-DIR SENDRECV SIZE:",I10,I10)')ibuff,wqf_buff(1)%size
  call comlib_sendrecv(wqf_buff(1)%id_sendup)
  call comlib_sendrecv(wqf_buff(1)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      qe%s(itht,iz,iy,1-NDEPTHX+ix) = TRANSFER(wqf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      qe%s(itht,iz,iy,     NX+1+ix) = TRANSFER(wqf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)

    enddo
  enddo
  enddo
  enddo
#endif

  return
end subroutine

function pack_u(u) result(pu)
  type(su3mat), intent(in) :: u
  complex(4) :: pu(VLEN,COL*(COL-1))
  integer :: ivec
  do ivec=1,VLEN
    pu(ivec,1) = CMPLX(u%u11r(ivec),u%u11i(ivec))
    pu(ivec,2) = CMPLX(u%u12r(ivec),u%u12i(ivec))
    pu(ivec,3) = CMPLX(u%u21r(ivec),u%u21i(ivec))
    pu(ivec,4) = CMPLX(u%u22r(ivec),u%u22i(ivec))
    pu(ivec,5) = CMPLX(u%u31r(ivec),u%u31i(ivec))
    pu(ivec,6) = CMPLX(u%u32r(ivec),u%u32i(ivec))
  enddo
  return
end function

function unpack_u(pu) result(u)
  complex(4), intent(in) :: pu(VLEN,COL*(COL-1))
  type(su3mat) :: u
  integer :: ivec
  do ivec=1,VLEN
    u%u11r(ivec) =  real(pu(ivec,1))
    u%u11i(ivec) = aimag(pu(ivec,1))
    u%u12r(ivec) =  real(pu(ivec,2))
    u%u12i(ivec) = aimag(pu(ivec,2))
    u%u21r(ivec) =  real(pu(ivec,3))
    u%u21i(ivec) = aimag(pu(ivec,3))
    u%u22r(ivec) =  real(pu(ivec,4))
    u%u22i(ivec) = aimag(pu(ivec,4))
    u%u31r(ivec) =  real(pu(ivec,5))
    u%u31i(ivec) = aimag(pu(ivec,5))
    u%u32r(ivec) =  real(pu(ivec,6))
    u%u32i(ivec) = aimag(pu(ivec,6))
  enddo
  return
end function

subroutine copy_boundary_ras_gvf_eo(ue)
  type(mic_ras_gvf_eo_soa), intent(inout) :: ue
  integer :: mu,ix,iy,iz,itht,ivec,ic,jc,ibuff
  integer, parameter :: NDAT = VLEN*COL*(COL-1)
  complex(4) :: ctmp(NDAT)
  type(su3mat) :: su3tmp

#if _NDIMZ != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1 + NX*(mu-1)))))
      gvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,NZ-NDEPTHZ+1+iz,iy,ix)%v(mu),mold=ctmp)
      gvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,           1+iz,iy,ix)%v(mu),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("U:Z-DIR SENDRECV SIZE:",I10,I10)')ibuff,gvf_buff(3)%size
  call comlib_sendrecv(gvf_buff(3)%id_sendup)
  call comlib_sendrecv(gvf_buff(3)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1 + NX*(mu-1)))))
      ue%s(itht,1-NDEPTHZ+iz,iy,ix)%v(mu) = TRANSFER(gvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      ue%s(itht,     NZ+1+iz,iy,ix)%v(mu) = TRANSFER(gvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
#endif
  
#if _NDIMY != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1 + NX*(mu-1)))))
      gvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,NY-NDEPTHY+1+iy,ix)%v(mu),mold=ctmp)
      gvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,           1+iy,ix)%v(mu),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("U:Y-DIR SENDRECV SIZE:",I10,I10)')ibuff,gvf_buff(2)%size
  call comlib_sendrecv(gvf_buff(2)%id_sendup)
  call comlib_sendrecv(gvf_buff(2)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1 + NX*(mu-1)))))
      ue%s(itht,iz,1-NDEPTHY+iy,ix)%v(mu) = TRANSFER(gvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      ue%s(itht,iz,     NY+1+iy,ix)%v(mu) = TRANSFER(gvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
#endif

#if _NDIMX != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*(ix + NDEPTHX*(mu-1)))))
      gvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,iy,NX-NDEPTHX+1+ix)%v(mu),mold=ctmp)
      gvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(ue%s(itht,iz,iy,           1+ix)%v(mu),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("U:X-DIR SENDRECV SIZE:",I10,I10)')ibuff,gvf_buff(1)%size
  call comlib_sendrecv(gvf_buff(1)%id_sendup)
  call comlib_sendrecv(gvf_buff(1)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(mu,ix,iy,iz,itht,ibuff,su3tmp) COLLAPSE(4)
  do mu=1,NDIM
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*(ix + NDEPTHX*(mu-1)))))
      ue%s(itht,iz,iy,1-NDEPTHX+ix)%v(mu) = TRANSFER(gvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
      ue%s(itht,iz,iy,     NX+1+ix)%v(mu) = TRANSFER(gvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=su3tmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
  enddo
#endif

  return
end subroutine

function pack_clv(f) result(pf)
  type(half_clover_term), intent(in) :: f
  complex(4) :: pf(VLEN,CLSPH-COL)
  integer :: ivec
  do ivec = 1,VLEN
    pf(ivec, 1) = CMPLX( f%f1r(ivec),f%f16r(ivec))
    pf(ivec, 2) = CMPLX( f%f7r(ivec),f%f19r(ivec))
    pf(ivec, 3) = CMPLX(f%f12r(ivec),f%f21r(ivec))
    pf(ivec, 4) = CMPLX( f%f2r(ivec), f%f2i(ivec))
    pf(ivec, 5) = CMPLX( f%f3r(ivec), f%f3i(ivec))
    pf(ivec, 6) = CMPLX( f%f4r(ivec), f%f4i(ivec))
    pf(ivec, 7) = CMPLX( f%f5r(ivec), f%f5i(ivec))
    pf(ivec, 8) = CMPLX( f%f6r(ivec), f%f6i(ivec))
    pf(ivec, 9) = CMPLX( f%f8r(ivec), f%f8i(ivec))
    pf(ivec,10) = CMPLX( f%f9r(ivec), f%f9i(ivec))
    pf(ivec,11) = CMPLX(f%f10r(ivec),f%f10i(ivec))
    pf(ivec,12) = CMPLX(f%f11r(ivec),f%f11i(ivec))
    pf(ivec,13) = CMPLX(f%f13r(ivec),f%f13i(ivec))
    pf(ivec,14) = CMPLX(f%f14r(ivec),f%f14i(ivec))
    pf(ivec,15) = CMPLX(f%f15r(ivec),f%f15i(ivec))
    pf(ivec,16) = CMPLX(f%f17r(ivec),f%f17i(ivec))
    pf(ivec,17) = CMPLX(f%f18r(ivec),f%f18i(ivec))
    pf(ivec,18) = CMPLX(f%f20r(ivec),f%f20i(ivec))
  enddo
  return
end function

function unpack_clv(pf) result(f)
  complex(4), intent(in) :: pf(VLEN,CLSPH-COL)
  type(half_clover_term) :: f
  integer :: ivec
  do ivec = 1,VLEN
     f%f1r(ivec) =  real(pf(ivec, 1))
    f%f16r(ivec) = aimag(pf(ivec, 1))
     f%f7r(ivec) =  real(pf(ivec, 2))
    f%f19r(ivec) = aimag(pf(ivec, 2))
    f%f12r(ivec) =  real(pf(ivec, 3))
    f%f21r(ivec) = aimag(pf(ivec, 3))
     f%f2r(ivec) =  real(pf(ivec, 4))
     f%f2i(ivec) = aimag(pf(ivec, 4))
     f%f3r(ivec) =  real(pf(ivec, 5))
     f%f3i(ivec) = aimag(pf(ivec, 5))
     f%f4r(ivec) =  real(pf(ivec, 6))
     f%f4i(ivec) = aimag(pf(ivec, 6))
     f%f5r(ivec) =  real(pf(ivec, 7))
     f%f5i(ivec) = aimag(pf(ivec, 7))
     f%f6r(ivec) =  real(pf(ivec, 8))
     f%f6i(ivec) = aimag(pf(ivec, 8))
     f%f8r(ivec) =  real(pf(ivec, 9))
     f%f8i(ivec) = aimag(pf(ivec, 9))
     f%f9r(ivec) =  real(pf(ivec,10))
     f%f9i(ivec) = aimag(pf(ivec,10))
    f%f10r(ivec) =  real(pf(ivec,11))
    f%f10i(ivec) = aimag(pf(ivec,11))
    f%f11r(ivec) =  real(pf(ivec,12))
    f%f11i(ivec) = aimag(pf(ivec,12))
    f%f13r(ivec) =  real(pf(ivec,13))
    f%f13i(ivec) = aimag(pf(ivec,13))
    f%f14r(ivec) =  real(pf(ivec,14))
    f%f14i(ivec) = aimag(pf(ivec,14))
    f%f15r(ivec) =  real(pf(ivec,15))
    f%f15i(ivec) = aimag(pf(ivec,15))
    f%f17r(ivec) =  real(pf(ivec,16))
    f%f17i(ivec) = aimag(pf(ivec,16))
    f%f18r(ivec) =  real(pf(ivec,17))
    f%f18i(ivec) = aimag(pf(ivec,17))
    f%f20r(ivec) =  real(pf(ivec,18))
    f%f20i(ivec) = aimag(pf(ivec,18))
  enddo
  return
end function

subroutine copy_boundary_ras_clvf_eo(fclve)
  type(mic_ras_clvf_eo), intent(inout) :: fclve
  integer :: mu,ix,iy,iz,itht,ivec,is,ibuff,ic
  integer, parameter :: NDAT=VLEN*(CLSPH-COL)*2
  complex(4) :: ctmp(NDAT)
  type(mic_clover) :: clvtmp

#if _NDIMZ != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      clvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,NZ-NDEPTHZ+1+iz,iy,ix),mold=ctmp)
      clvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,           1+iz,iy,ix),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("CLV:Z-DIR SENDRECV SIZE:",I10,I10)')ibuff,clvf_buff(3)%size
  call comlib_sendrecv(clvf_buff(3)%id_sendup)
  call comlib_sendrecv(clvf_buff(3)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,clvtmp) COLLAPSE(3)
  do ix=1,NX
  do iy=1,NY
  do iz=0,NDEPTHZ-1
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz + NDEPTHZ*(iy-1 + NY*(ix-1))))
      fclve%s(itht,1-NDEPTHZ+iz,iy,ix) = TRANSFER(clvf_buff(3)%buff_up(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
      fclve%s(itht,     NZ+1+iz,iy,ix) = TRANSFER(clvf_buff(3)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
#endif
  
#if _NDIMY != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      clvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,NY-NDEPTHY+1+iy,ix),mold=ctmp)
      clvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,           1+iy,ix),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("CLV:Y-DIR SENDRECV SIZE:",I10,I10)')ibuff,clvf_buff(2)%size
  call comlib_sendrecv(clvf_buff(2)%id_sendup)
  call comlib_sendrecv(clvf_buff(2)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,clvtmp) COLLAPSE(3)
  do ix=1,NX
  do iy=0,NDEPTHY-1
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy + NDEPTHY*(ix-1))))
      fclve%s(itht,iz,1-NDEPTHY+iy,ix) = TRANSFER(clvf_buff(2)%buff_up(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
      fclve%s(itht,iz,     NY+1+iy,ix) = TRANSFER(clvf_buff(2)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
#endif

#if _NDIMX != 1
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,ctmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      clvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,iy,NX-NDEPTHX+1+ix),mold=ctmp)
      clvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,1) = TRANSFER(fclve%s(itht,iz,iy,           1+ix),mold=ctmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
!  if (0 == nodeid) write(*,'("CLV:X-DIR SENDRECV SIZE:",I10,I10)')ibuff,clvf_buff(1)%size
  call comlib_sendrecv(clvf_buff(1)%id_sendup)
  call comlib_sendrecv(clvf_buff(1)%id_senddn)
  ibuff = 0
!$OMP PARALLEL DO PRIVATE(ix,iy,iz,itht,ibuff,clvtmp) COLLAPSE(3)
  do ix=0,NDEPTHX-1
  do iy=1-NDEPTHY,NY+NDEPTHY
  do iz=1-NDEPTHZ,NZ+NDEPTHZ
    do itht=1,NTHT

      ibuff = NDAT*(itht-1 + NTHT*(iz-1+NDEPTHZ + (NZ+NDEPTHZ*2)*(iy-1+NDEPTHY + (NY+NDEPTHY*2)*ix)))
      fclve%s(itht,iz,iy,1-NDEPTHX+ix) = TRANSFER(clvf_buff(1)%buff_up(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
      fclve%s(itht,iz,iy,     NX+1+ix) = TRANSFER(clvf_buff(1)%buff_dn(ibuff+1:ibuff+NDAT,2),mold=clvtmp)
!      ibuff = ibuff + NDAT

    enddo
  enddo
  enddo
  enddo
#endif

  return
end subroutine

end module
