#ifndef _MIC_WRAPPER_FUNC_H
#define _MIC_WRAPPER_FUNC_H

#if __MIC__ == 1

#define _ALIGN_SIZE               (64)

#define _my_setzero_ps()          _mm512_set1_ps(0.0f)

#define _my_set1_ps(a)            _mm512_set1_ps((a))
#define _my_load_ps(a)            _mm512_load_ps((void *)(a))
#define _my_store_ps(a,b)         _mm512_store_ps((void *)(a),(b))

#define _my_ngostore_ps(a,b)      _mm512_storenrngo_ps((void *)(a),(b))
#define _my_pf(a,b)               _mm_prefetch((char *)(a),(b))

#define _my_load_bcast_ss(a)      _mm512_extload_ps((void *)(a),_MM_UPCONV_PS_NONE,_MM_BROADCAST_1X16,_MM_HINT_NONE)

#define _my_mul_ps(a,b)           _mm512_mul_ps((a),(b))
#define _my_sub_ps(a,b)           _mm512_sub_ps((a),(b))
#define _my_add_ps(a,b)           _mm512_add_ps((a),(b))
#define _my_sub_ps(a,b)           _mm512_sub_ps((a),(b))
#define _my_fmadd_ps(a,b,c)       _mm512_fmadd_ps((a),(b),(c))
#define _my_fmsub_ps(a,b,c)       _mm512_fmsub_ps((a),(b),(c))
#define _my_fnmadd_ps(a,b,c)      _mm512_fnmadd_ps((a),(b),(c))
#define _my_fnmsub_ps(a,b,c)      _mm512_fnmsub_ps((a),(b),(c))

#define _my_xor_ps(a,mask)        (__m512)_mm512_mask_xor_epi32((__m512i)(a),(mask),(__m512i)(a),(__m512i)_mm512_set1_ps(-0.0f))

#define _my_reduce_add_ps(a)      _mm512_reduce_add_ps((a))

#define _my_swap_reim_ps(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_BADC)
#define _my_swap_spin_ps(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_CDAB)
#define _my_swap_spin_reim_ps(a)  (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_ABCD)

#define _my_permute_BADC(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_BADC)
#define _my_permute_CDAB(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_CDAB)
#define _my_permute_CBAD(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_CBAD)
#define _my_permute_ADCB(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_ADCB)
#define _my_permute_ABCD(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_ABCD)
#define _my_permute_AAAA(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_AAAA)
#define _my_permute_BBBB(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_BBBB)
#define _my_permute_CCCC(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_CCCC)
#define _my_permute_DDDD(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_DDDD)
#define _my_permute_BABA(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_BABA)
#define _my_permute_ABAB(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_ABAB)
#define _my_permute_DCDC(a)       (__m512)_mm512_permute4f128_epi32((__m512i)(a),_MM_PERM_DCDC)


#elif __AVX__ == 1

#define _ALIGN_SIZE               (32)

#define _my_setzero_ps()          _mm256_setzero_ps()

#define _my_set1_ps(a)            _mm256_set1_ps((a))
#define _my_load_ps(a)            _mm256_load_ps((float *)(a))
#define _my_store_ps(a,b)         _mm256_store_ps((float *)(a),(b))

#define _my_load_bcast_ss(a)      _mm256_broadcast_ss((float *)(a))

#define _my_mul_ps(a,b)           _mm256_mul_ps((a),(b))
#define _my_add_ps(a,b)           _mm256_add_ps((a),(b))
#define _my_sub_ps(a,b)           _mm256_sub_ps((a),(b))

// a*b + c
#define _my_fmadd_ps(a,b,c)       _mm256_add_ps(_mm256_mul_ps((a),(b)),(c));

// a*b - c
#define _my_fmsub_ps(a,b,c)       _mm256_sub_ps(_mm256_mul_ps((a),(b)),(c));

// - a*b + c
#define _my_fnmadd_ps(a,b,c)      _mm256_xor_ps(_mm256_sub_ps(_mm256_mul_ps((a),(b)),(c)),_mm256_set1_ps(-0.0f));

// - a*b - c
#define _my_fnmsub_ps(a,b,c)      _mm256_xor_ps(_mm256_add_ps(_mm256_mul_ps((a),(b)),(c)),_mm256_set1_ps(-0.0f));

#define _my_xor_ps(a,mask)        _mm256_xor_ps((a),(mask))

inline float _my_reduce_add_ps(__m256 a) 
{
  register __m256 btmp = _mm256_hadd_ps(a,a);
  btmp = _mm256_hadd_ps(btmp,btmp);
  btmp = _mm256_add_ps(btmp,_mm256_permute2f128_ps(btmp,btmp,0x01));
  __m128 atmp = _mm256_castps256_ps128(btmp);
  float aa;
  _mm_store_ss(&aa,atmp);
  return aa;
/*
  float aa[8];
  _mm256_store_ps(aa,btmp);
  return aa[0];
*/
}

#define _my_swap_reim_ps(a)       _mm256_permute2f128_ps((a),(a),0x01)
#define _my_swap_spin_ps(a)       _mm256_permute_ps((a),_MM_PERM_BADC)
#define _my_swap_spin_reim_ps(a)  _mm256_permute_ps(_mm256_permute2f128_ps((a),(a),0x01),_MM_PERM_BADC)

#define _my_permute_BADC(a)       _mm256_permute2f128_ps((a),(a),0x01)
#define _my_permute_CDAB(a)       _mm256_permute_ps((a),_MM_PERM_BADC)

inline __m256 _my_permute_CBAD(__m256 a) 
{
  register __m256 btmp = _mm256_permute_ps((a),_MM_PERM_BADC);
  register __m256 ctmp = _mm256_permute2f128_ps(btmp,btmp,0x01);
  return _mm256_blend_ps(btmp,ctmp,0x33);
}

inline __m256 _my_permute_ADCB(__m256 a) 
{
  register __m256 btmp = _mm256_permute_ps((a),_MM_PERM_BADC);
  register __m256 ctmp = _mm256_permute2f128_ps(btmp,btmp,0x01);
  return _mm256_blend_ps(ctmp,btmp,0x33);
}

#define _my_permute_ABCD(a)      _mm256_permute_ps(_mm256_permute2f128_ps((a),(a),0x01),_MM_PERM_BADC)

inline __m256 _my_permute_AAAA(__m256 a) 
{
  register __m256 btmp = _mm256_permute2f128_ps((a),(a),0x00);
  register __m256 ctmp = _mm256_permute_ps(btmp,_MM_PERM_BADC);
  return _mm256_blend_ps(ctmp,btmp,0x33);
}

inline __m256 _my_permute_BBBB(__m256 a) 
{
  register __m256 btmp = _mm256_permute2f128_ps((a),(a),0x00);
  register __m256 ctmp = _mm256_permute_ps(btmp,_MM_PERM_BADC);
  return _mm256_blend_ps(btmp,ctmp,0x33);
}

inline __m256 _my_permute_CCCC(__m256 a) 
{
  register __m256 btmp = _mm256_permute2f128_ps((a),(a),0x11);
  register __m256 ctmp = _mm256_permute_ps(btmp,_MM_PERM_BADC);
  return _mm256_blend_ps(ctmp,btmp,0x33);
}

inline __m256 _my_permute_DDDD(__m256 a) 
{
  register __m256 btmp = _mm256_permute2f128_ps((a),(a),0x11);
  register __m256 ctmp = _mm256_permute_ps(btmp,_MM_PERM_BADC);
  return _mm256_blend_ps(btmp,ctmp,0x33);
}

#define _my_permute_BABA(a)      _mm256_permute2f128_ps((a),(a),0x00)
#define _my_permute_ABAB(a)      _mm256_permute_ps(_mm256_permute2f128_ps((a),(a),0x00),_MM_PERM_BADC)
#define _my_permute_DCDC(a)      _mm256_permute2f128_ps((a),(a),0x11)
#define _my_permute_CDCD(a)      _mm256_permute_ps(_mm256_permute2f128_ps((a),(a),0x11),_MM_PERM_BADC)

#elif __SSE4_1__ == 1

#define _ALIGN_SIZE               (16)

#define _my_setzero_ps()          _mm_setzero_ps()
#define _my_set1_ps(a)            _mm_set1_ps((a))
#define _my_load_ps(a)            _mm_load_ps((float *)(a))
#define _my_store_ps(a,b)         _mm_store_ps((float *)(a),(b))

#define _my_load_bcast_ss(a)      _mm_load1_ps((float *)(a))
#define _my_mul_ps(a,b)           _mm_mul_ps((a),(b))
#define _my_add_ps(a,b)           _mm_add_ps((a),(b))
#define _my_sub_ps(a,b)           _mm_sub_ps((a),(b))

// a*b + c
#define _my_fmadd_ps(a,b,c)       _mm_add_ps(_mm_mul_ps((a),(b)),(c));

// a*b - c
#define _my_fmsub_ps(a,b,c)       _mm_sub_ps(_mm_mul_ps((a),(b)),(c));

// - a*b + c
#define _my_fnmadd_ps(a,b,c)      _mm_xor_ps(_mm_sub_ps(_mm256_mul_ps((a),(b)),(c)),_mm_set1_ps(-0.0f));

// - a*b - c
#define _my_fnmsub_ps(a,b,c)      _mm_xor_ps(_mm_add_ps(_mm256_mul_ps((a),(b)),(c)),_mm_set1_ps(-0.0f));

#define _my_xor_ps(a,mask)        _mm_xor_ps((a),(mask))

inline float _my_reduce_add_ps(__m128 a) 
{
  register __m128 atmp = _mm_dp_ps((a),(a),0xFF);
  _ALIG_PRE_ float aa _ALIG_PST_;
  _mm_store_ss(&aa,atmp);
  return aa;
}

//    ( y2i, y1i, y2r, y1r) =>  ( y2r, y1r, y2i, y1i)
#define _my_swap_reim_ps(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(1,0,3,2))

//    ( y2i, y1i, y2r, y1r) =>  ( y1i, y2i, y1r, y2r)
#define _my_swap_spin_ps(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(2,3,0,1))

//    ( y2i, y1i, y2r, y1r) =>  ( y1r, y2r, y1i, y2i)
#define _my_swap_spin_reim_ps(a)  _mm_shuffle_ps((a),(a),_MM_SHUFFLE(0,1,2,3))

//    DCBA => BADC
#define _my_permute_BADC(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(1,0,3,2))

//    DCBA => CDAB
#define _my_permute_CDAB(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(2,3,0,1))

#define _my_permute_CBAD(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(2,1,0,3))
#define _my_permute_ADCB(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(0,3,2,1))
#define _my_permute_ABCD(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(0,1,2,3))

#define _my_permute_AAAA(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(0,0,0,0))
#define _my_permute_BBBB(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(1,1,1,1))
#define _my_permute_CCCC(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(2,2,2,2))
#define _my_permute_DDDD(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(3,3,3,3))

#define _my_permute_BABA(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(1,0,1,0))
#define _my_permute_ABAB(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(0,1,0,1))
#define _my_permute_DCDC(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(3,2,3,2))
#define _my_permute_CDCD(a)       _mm_shuffle_ps((a),(a),_MM_SHUFFLE(2,3,2,3))

#endif

#endif
