#ifndef __SCIF_INTERFACE_H__
#define __SCIF_INTERFACE_H__

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <scif.h>

#define BACKLOG 16

struct Command
{
    uint32_t cmd;
    uint32_t spare;
    uint64_t cmdlen;
    uint64_t data[];
    //#pragma offload_attribute target(mic)
    static const int CMD_Stop = 1;
};

class SCIF_Interface {
	
  public:
    // Constructor
    SCIF_Interface();
    // Create active endpoint (host) and connect to (node,port)
    //#pragma offload_attribute target(mic)
    void create_send_port(int node, uint16_t port);
    // Create passive endpoint (card) and return port and node
    //#pragma offload_attribute target(mic)
    uint16_t create_receive_port(int &node);
    // Wait for connection (card)
    //#pragma offload_attribute target(mic)
    void server();
    // Send message in given buffer
    //#pragma offload_attribute target(mic)
    void send_message(char *buffer,unsigned msg_size, bool block);
    // Receive message in given buffer
    //#pragma offload_attribute target(mic)
    void receive_message(char *buffer, unsigned msg_size, bool block);
    // Get pointer to the send buffer
    char *get_send_buffer(void);
    // Register a buffer and return offset into new addr. space
    off_t register_buffer(char *buffer, unsigned buffer_size, bool read_only);
    // mmap a remote buffer at offset off and return local address
    void *mmap_buffer(size_t buffer_size, off_t off, bool read_only);
    // copy to remote buffer
    void write(off_t local_offset, off_t remote_offset, unsigned buffer_size);
    // copy from remote buffer
    void read(off_t local_offset, off_t remote_offset, unsigned buffer_size);
    // Wait until all outstanding RMAs are done
    void fence(bool self_fence=true);
    void close(void); // Close the driver
    // get node IDs in system
    //#pragma offload_attribute target(mic)
    unsigned get_nodes(uint16_t *nodes, int len, uint16_t *self);
    off_t init_asynch_fence(void);
    //#pragma offload_attribute target(mic)
    void asynch_fence(off_t offset, unsigned value);
    //#pragma offload_attribute target(mic)
    inline bool asynch_poll(unsigned value)
    {
        return (*((volatile unsigned *)signal_addr) == value);
    }
    //#pragma offload_attribute target(mic)
    inline void clear_asynch_signal(void)
    {
        *((volatile unsigned *)signal_addr) = 0;
    }
    // dequeue a message. Return true if stop message not received.
    // If len is 0, no message is available, otherwise data points to the
    // message data
    void dequeue(int &data, uint64_t &msgstart, size_t &len, bool block);
    // copy the dequeued message
    void qcopy(Command *msg, uint64_t msgstart, size_t len);
    // enqueue a message. Returns true on success, false if no space
    //#pragma offload_attribute target(mic)
    bool enqueue(Command *msg, size_t len, bool block);
    // check if queue is empty
    //#pragma offload_attribute target(mic)
    bool qempty(bool block);
    // check EP for status
    //#pragma offload_attribute target(mic)
    void checkep();
    // registered buffers.
    // Allocate a new buffer and return offset (card only)
    bool allocbuf(int len, off_t &off, void *&buf);
    // Get the buffer address. off is the card-side offset
    void *getbuf(off_t off);
    // Read a card buffer into the corresponding buffer on the host
    void readbuf(off_t off, size_t len);
    // Write a card buffer from the corresponding buffer on the host
    void writebuf(off_t off, size_t len);
    // reset the buffers
    void clearbuf() { rcur = 0; }

    void *signal_addr;

  private:
    scif_epd_t epd;
    scif_epd_t newepd;
    scif_epd_t tmpepd;
    int sciferr;
    int con_pn;
    struct scif_portID portID;
    int msg_size;
    int block;
    int node;
    bool sender;
    const int msgsize = 64;

    // message queue. One side sends, other side recieves
    const int qlen = 4096;//32*1024;
    char *qbuf;
    off_t qoff;         // local

    // buffers. Each side allocates and registers. Card also sends local offset
    // to host
    const int rbufz = 64*1024*1024;
    char *rbuf;
    off_t roff;
    off_t roffrem;      // host only, on card same as roff
    int rcur;   // current free position in rbuf, card only
};

#endif
