#include "config.h"

program ccs_qcd_solver_bench
!***********************************************************************
!$Id: ccs_qcd_solver_bench.F90,v 1.4 2012/12/19 02:17:33 ishikawa Exp $
!
!  CCS QCD solver benchmark test program.
!
!  This benchmark test program measures CPU and MPI performances.
!
!***********************************************************************
  use comlib
  use lattice_class
  use ccs_qcd_solver_bench_class
  implicit none
!***********************************************************************
  complex(8) ::  ye(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(8) ::  yo(COL,SPIN,0:NTH,0:NZ1,0:NY1,0:NX1)
  complex(8) ::  ue(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)
  complex(8) ::  uo(COL,COL,0:NTH,0:NZ1,0:NY1,0:NX1,NDIM)

  integer :: iargc, iflag

  character(256) :: arg

  integer :: iloop,NLOOP
  integer(8) :: config_size
  real(8) :: etime0,etime1
  real(8) :: write_etime, write_size
  real(8) :: read_etime,   read_size
  integer, parameter :: iout=_FILE_IO_BICGSTAB_HMC

  config_size = COL*COL*(NTH+1)*(NZ+2)*(NY+2)*(NX+2)*NDIM*2*16  ! byte

  call initset

  mult_iter = 0
  write_etime = 0.0d0
   read_etime = 0.0d0

  if (iargc() /= 2) then
    if (nodeid==0) then
      write(*,*)
      write(*,'("Usage: ccs_qcd_io_bench loop_count flag")')
      write(*,*)
      write(*,'("CCS QCD Input-Output configuration Benchmark Program:")')
      write(*,'("Write or Read QCD configuration")')
      write(*,*)
      write(*,'(" Command args:")')
      write(*,'(" loop_count       Execution count for BiCGStab solver. (integer number)")')
      write(*,'(" flag             0 for config write, 1 for config read.")')
      write(*,'(80("="))')
      write(*,'(" CCS QCD I/O Benchmark Program")')
      write(*,'(X,A)')_VERSION_
      write(*,'(X,A)')_REVISION_
      write(*,'(80("="))')
    endif
#ifndef _singlePU
    call comlib_finalize
#endif
    stop
  endif
  if (nodeid==0) then
    call getarg(1,arg)
    read(arg,*)NLOOP
    call getarg(2,arg)
    read(arg,*)iflag
  endif
#ifndef _singlePU
  call comlib_bcast(NLOOP,0)
  call comlib_bcast(iflag,0)
#endif

  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'(" CCS QCD I/O Benchmark Program")')
    write(*,'(X,A)')_VERSION_
    write(*,'("    Lattice Size :",3I3,I4)')NTX,NTY,NTZ,NTT
    write(*,'(" Node Array Size :",3I3)')NDIMX,NDIMY,NDIMZ
    write(*,'(" PU Lattice Size :",3I3,I4,"*2")')NX,NY,NZ,NTH
    write(*,'("            Loop :",I9)')NLOOP
    select case(iflag)
    case(0)
    write(*,'("    Write Config :")')
    case(1)
    write(*,'("     Read Config :")')
    end select
    write(*,'(80("="))')
  endif

!***************************
! initialize
! (ue,uo) and (ye,yo)
! with Gaussian rand. num.
!***************************
  call init_u_and_y(ue,uo,ye,yo)

  call output(ue,uo,ye,yo)

  select case(iflag)
  case(0)

    do iloop=1,NLOOP
      call xclock(etime0,8)
      call write_config(ue,uo,iout,iloop)
      call xclock(etime1,8)
      write_etime = write_etime + (etime1-etime0)
    enddo

  case(1)

    do iloop=1,NLOOP
      call xclock(etime0,8)
      call read_config(ue,uo,iout,iloop)
      call xclock(etime1,8)
      read_etime = read_etime + (etime1-etime0)
    enddo

  end select

  call output(ue,uo,ye,yo)

  write_size = REAL(config_size,kind=KIND(1.0d0))*NPU * NLOOP
  read_size  = write_size

!******************
! write results
!******************
  if (nodeid==0) then
    write(*,'(80("="))')
    write(*,'("  QCD config write Performance.")')
    select case(iflag)
    case(0)
      write(*,'("            Write Time [sec] :",32(F12.4))') write_etime
      write(*,'("          Write Size [Bytes] :",32(E14.6))') write_size
      write(*,'("     Performance [Bytes/sec] :",32(E14.6))') write_size/write_etime
    case(1)
      write(*,'("             Read Time [sec] :",32(F12.4))') read_etime
      write(*,'("           Read Size [Bytes] :",32(E14.6))') read_size
      write(*,'("     Performance [Bytes/sec] :",32(E14.6))') read_size/read_etime
    end select

    write(*,'(80("="))')
  endif

#ifndef _singlePU
  call comlib_finalize
#endif

  stop
end program
